!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-------------------------------------------------------------------------------
!     ##############################
      MODULE MODI_ADVEC_WENO_K_2_AUX
!     ##############################
!
INTERFACE
!
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_UX(HLBCX,PSRC, PRUCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_UX(HLBCX,PSRC, PRUCT, PR, TWEST, TEAST,            &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX  ! X direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRUCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRUCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TWEST,TEAST
!$acc declare copyin(TWEST,TEAST)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
#endif
!
END SUBROUTINE ADVEC_WENO_K_2_UX
!
!                    ----------------------------
!
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_MX(HLBCX,PSRC, PRUCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_MX(HLBCX,PSRC, PRUCT, PR, TWEST, TEAST,            &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX  ! X direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRUCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRUCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TWEST,TEAST
!$acc declare copyin(TWEST,TEAST)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
#endif
!
END SUBROUTINE ADVEC_WENO_K_2_MX
!
!                     ---------------------------
!
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_VY(HLBCY,PSRC, PRVCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_VY(HLBCY,PSRC, PRVCT, PR, TNORTH, TSOUTH,          &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY  ! Y direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRVCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRVCT)
!
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TNORTH,TSOUTH
!$acc declare copyin(TNORTH,TSOUTH)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
#endif
!
END SUBROUTINE ADVEC_WENO_K_2_VY
!
!                  ------------------------------
!
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_MY(HLBCY,PSRC, PRVCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_MY(HLBCY,PSRC, PRVCT, PR, TNORTH, TSOUTH,          &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY  ! Y direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRVCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRVCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TNORTH,TSOUTH
!$acc declare copyin(TNORTH,TSOUTH)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
#endif
!
END SUBROUTINE ADVEC_WENO_K_2_MY
!
!                     -------------------------------
!
#ifndef _OPENACC
FUNCTION WENO_K_2_WZ(PSRC, PRWCT) RESULT(PR)
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on W grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRWCT)
!
! output source term
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: PR
!
END FUNCTION WENO_K_2_WZ
#else
SUBROUTINE WENO_K_2_WZ(PSRC, PRWCT, PR,                                &
                       ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                       ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on W grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRWCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
END SUBROUTINE WENO_K_2_WZ
#endif
!
!                      ------------------------------
!
#ifndef _OPENACC
FUNCTION WENO_K_2_MZ(PSRC, PRWCT) RESULT(PR)
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on MASS grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on W grid
!$acc declare present(PSRC,PRWCT)
!
! output source term
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: PR
!
END FUNCTION WENO_K_2_MZ
#else
SUBROUTINE WENO_K_2_MZ(PSRC, PRWCT, PR,                                &
                       ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                       ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on MASS grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on W grid
!$acc declare present(PSRC,PRWCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
!
! Work arrays
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFPOS1,ZFPOS2,ZFNEG1,ZFNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
END SUBROUTINE WENO_K_2_MZ
#endif
!
END INTERFACE
!
END MODULE MODI_ADVEC_WENO_K_2_AUX
!
!-----------------------------------------------------------------------------
!
!     ############################################################
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_UX(HLBCX,PSRC, PRUCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_UX(HLBCX,PSRC, PRUCT, PR, TWEST, TEAST,            &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     ############################################################
!!
!!**** Computes PRUCT * PUT. Upstream fluxes of U in X direction.  
!!     Input PUT is on U Grid 'ie' (i,j,k) based on UGRID reference
!!     Output PR is on mass Grid 'ie' (i+1/2,j,k) based on UGRID reference
!!              
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*               
!!
!!    MODIFICATIONS
!!    -------------
!!      J.Escobar 21/03/2013: for HALOK comment all NHALO=1 test
!!
!-------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODI_GET_HALO
USE MODE_MPPDB
#ifdef _OPENACC
use mode_msg
#endif
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX  ! X direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRUCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRUCT)
!
! output source term
!
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
REAL, DIMENSION(:,:), POINTER :: TWEST, TEAST
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TWEST,TEAST
!$acc declare copyin(TWEST,TEAST)
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IIB,IJB    ! Begining useful area in x,y,z directions
INTEGER :: IIE,IJE    ! End useful area in x,y,z directions
INTEGER:: IW,IE,IWF,IEF   ! Coordinate of third order diffusion area
!
INTEGER:: ILUOUT,IRESP   ! for prints
!
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!-----------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "ADVEC_WENO_K_2_UX beg:PSRC")
  CALL MPPDB_CHECK(PRUCT,"ADVEC_WENO_K_2_UX beg:PRUCT")
END IF
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
!
!-------------------------------------------------------------------------------
!
!*       0.4.   INITIALIZE THE FIELD 
!               ---------------------
!
#ifndef _OPENACC
IF (PRESENT(TPHALO2)) THEN
  TWEST => TPHALO2%WEST
  TEAST => TPHALO2%EAST
END IF
#endif
!
!$acc kernels
PR(:,:,:) = 0.0
!
ZFPOS1  = 0.0
ZFPOS2  = 0.0
ZFNEG1  = 0.0
ZFNEG2  = 0.0
ZBPOS1  = 0.0
ZBPOS2  = 0.0
ZBNEG1  = 0.0
ZBNEG2  = 0.0
ZOMP1   = 0.0
ZOMP2   = 0.0
ZOMN1   = 0.0
ZOMN2   = 0.0
!
!-------------------------------------------------------------------------------
!
SELECT CASE ( HLBCX(1) ) ! X direction LBC type: (1) for left side
!
!*       1.1    CYCLIC CASE IN THE X DIRECTION:
!
CASE ('CYCL')          ! In that case one must have HLBCX(1) == HLBCX(2)
#ifdef _OPENACC
  call Print_msg( NVERB_WARNING, 'GEN', 'ADVEC_WENO_K_2_UX', 'OpenACC: CYCL not yet tested' )
#endif
!
  IW=IIB
  IE=IIE
!
! r: many left cells in regard to 'i' cell for each stencil
!
! intermediate fluxes at the mass point on Ugrid u(i+1/2,j,k) for positive wind
! (r=1 for the first stencil ZFPOS1, r=0 for the second ZFPOS2)
!
   ZFPOS1(IW:IE+1,:,:) = 0.5 * (3.0*PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))
   ZFPOS1(IW-1,   :,:) = 0.5 * (3.0*PSRC(IW-1,   :,:) - TWEST(:,:))
!
   ZFPOS2(IW-1:IE,:,:) = 0.5 * (PSRC(IW-1:IE,:,:) + PSRC(IW:IE+1,:,:))
   ZFPOS2(IE+1,   :,:) = 0.5 * (PSRC(IE+1,   :,:) + TEAST(:,:))
!
! intermediate flux at the mass point on Ugrid (i+1/2,j,k) for negative wind 
! case (from the right to the left)
! (r=0 for the second stencil ZFNEG2=ZFPOS2, r=-1 for the first ZFNEG1)  
!
  ZFNEG1(IW-1:IE-1,:,:) = 0.5 * (3.0*PSRC(IW:IE,:,:)   - PSRC(IW+1:IE+1,:,:))
  ZFNEG1(IE,   :,:)     = 0.5 * (3.0*PSRC(IE+1,   :,:) - TEAST(:,:))
  ZFNEG2(IW-1:IE,:,:)   = 0.5 * (PSRC(IW-1:IE,:,:) + PSRC(IW:IE+1,:,:))
  ZFNEG2(IE+1,   :,:)   = 0.5 * (PSRC(IE+1,   :,:) + TEAST(:,:))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
  ZBPOS1(IW:IE+1,:,:) = (PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))**2
  ZBPOS1(IW-1,   :,:) = (PSRC(IW-1,   :,:) - TWEST(:,:))**2
!
  ZBPOS2(IW-1:IE,:,:) = (PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))**2
  ZBPOS2(IE+1,   :,:) = (TEAST(:,:)        - PSRC(IE+1,   :,:))**2
!
! smoothness indicators for negative wind case
!
  ZBNEG1(IW-1:IE-1,:,:) = (PSRC(IW:IE,:,:)   - PSRC(IW+1:IE+1,:,:))**2
  ZBNEG1(IE,   :,:)     = (PSRC(IE+1,   :,:) - TEAST(:,:))**2
  ZBNEG2(IW-1:IE,:,:)   = (PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))**2
  ZBNEG2(IE+1,   :,:)   = (PSRC(IE+1,   :,:) - TEAST(:,:))**2
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / (ZEPS + ZBPOS1)**2
  ZOMP2 = ZGAMMA2 / (ZEPS + ZBPOS2)**2
  ZOMN1 = ZGAMMA1 / (ZEPS + ZBNEG1)**2
  ZOMN2 = ZGAMMA2 / (ZEPS + ZBNEG2)**2
#else
!
! smoothness indicators for positive wind case
!
  ZBPOS1(IW:IE+1,:,:) = BR_P2(PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))
  ZBPOS1(IW-1,   :,:) = BR_P2(PSRC(IW-1,   :,:) - TWEST(:,:))
!
  ZBPOS2(IW-1:IE,:,:) = BR_P2(PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))
  ZBPOS2(IE+1,   :,:) = BR_P2(TEAST(:,:)        - PSRC(IE+1,   :,:))
!
! smoothness indicators for negative wind case
!
  ZBNEG1(IW-1:IE-1,:,:) = BR_P2(PSRC(IW:IE,:,:)   - PSRC(IW+1:IE+1,:,:))
  ZBNEG1(IE,   :,:)     = BR_P2(PSRC(IE+1,   :,:) - TEAST(:,:))
  ZBNEG2(IW-1:IE,:,:)   = BR_P2(PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))
  ZBNEG2(IE+1,   :,:)   = BR_P2(PSRC(IE+1,   :,:) - TEAST(:,:))
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1)
  ZOMP2 = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2)
  ZOMN1 = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1)
  ZOMN2 = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2)
#endif
!
! WENO fluxes
!
  PR = (ZOMN2/(ZOMN1+ZOMN2) * ZFNEG2 +                           &
       (ZOMN1/(ZOMN1+ZOMN2) * ZFNEG1)) * (0.5-SIGN(0.5,PRUCT)) + &
       (ZOMP2/(ZOMP1+ZOMP2) * ZFPOS2 +                           &
       (ZOMP1/(ZOMP1+ZOMP2) * ZFPOS1)) * (0.5+SIGN(0.5,PRUCT))
!
!
!       OPEN, WALL, NEST CASE IN THE X DIRECTION
!
CASE ('OPEN','WALL','NEST')
!
  IW=IIB
  IE=IIE
!
!       USE A FIRST ORDER UPSTREAM SCHEME AT THE PHYSICAL BORDER
!
  IF(LWEST_ll()) THEN
    PR(IW-1,:,:) = PSRC(IW-1,:,:) * (0.5+SIGN(0.5,PRUCT(IW-1,:,:))) + PSRC(IW,:,:) * (0.5-SIGN(0.5,PRUCT(IW-1,:,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE 
#ifndef MNH_BITREP
    ZFPOS1(IW-1,:,:) = 0.5 * (3.0*PSRC(IW-1,:,:) - TWEST(:,:))
    ZFPOS2(IW-1,:,:) = 0.5 * (PSRC(IW-1,    :,:) + PSRC(IW,:,:))
    ZBPOS1(IW-1,:,:) = (PSRC(IW-1,:,:) - TWEST(:,:))**2
    ZBPOS2(IW-1,:,:) = (PSRC(IW,  :,:) - PSRC(IW-1,:,:))**2
!
    ZFNEG1(IW-1,:,:) = 0.5 * (3.0*PSRC(IW,:,:) - PSRC(IW+1,:,:))
    ZFNEG2(IW-1,:,:) = 0.5 * (PSRC(IW-1,  :,:) + PSRC(IW,  :,:))
    ZBNEG1(IW-1,:,:) = (PSRC(IW,  :,:) - PSRC(IW+1,:,:))**2
    ZBNEG2(IW-1,:,:) = (PSRC(IW-1,:,:) - PSRC(IW,  :,:))**2
!
    ZOMP1(IW-1,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IW-1,:,:))**2
    ZOMP2(IW-1,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IW-1,:,:))**2
    ZOMN1(IW-1,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IW-1,:,:))**2
    ZOMN2(IW-1,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IW-1,:,:))**2
#else
    ZFPOS1(IW-1,:,:) = 0.5 * (3.0*PSRC(IW-1,:,:) - TWEST(:,:))
    ZFPOS2(IW-1,:,:) = 0.5 * (PSRC(IW-1,    :,:) + PSRC(IW,:,:))
    ZBPOS1(IW-1,:,:) = BR_P2(PSRC(IW-1,:,:) - TWEST(:,:))
    ZBPOS2(IW-1,:,:) = BR_P2(PSRC(IW,  :,:) - PSRC(IW-1,:,:))
!
    ZFNEG1(IW-1,:,:) = 0.5 * (3.0*PSRC(IW,:,:) - PSRC(IW+1,:,:))
    ZFNEG2(IW-1,:,:) = 0.5 * (PSRC(IW-1,  :,:) + PSRC(IW,  :,:))
    ZBNEG1(IW-1,:,:) = BR_P2(PSRC(IW,  :,:) - PSRC(IW+1,:,:))
    ZBNEG2(IW-1,:,:) = BR_P2(PSRC(IW-1,:,:) - PSRC(IW,  :,:))
!
    ZOMP1(IW-1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IW-1,:,:))
    ZOMP2(IW-1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IW-1,:,:))
    ZOMN1(IW-1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IW-1,:,:))
    ZOMN2(IW-1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IW-1,:,:))
#endif
!
    PR(IW-1,:,:) = (ZOMN2(IW-1,:,:)/(ZOMN1(IW-1,:,:)+ZOMN2(IW-1,:,:)) * ZFNEG2(IW-1,:,:) +   &
                   (ZOMN1(IW-1,:,:)/(ZOMN1(IW-1,:,:)+ZOMN2(IW-1,:,:)) * ZFNEG1(IW-1,:,:))) * (0.5-SIGN(0.5,PRUCT(IW-1,:,:))) + &
                   (ZOMP2(IW-1,:,:)/(ZOMP1(IW-1,:,:)+ZOMP2(IW-1,:,:)) * ZFPOS2(IW-1,:,:) +  &
                   (ZOMP1(IW-1,:,:)/(ZOMP1(IW-1,:,:)+ZOMP2(IW-1,:,:)) * ZFPOS1(IW-1,:,:))) * (0.5+SIGN(0.5,PRUCT(IW-1,:,:)))
!
  ENDIF
!
  IF(LEAST_ll()) THEN
    PR(IE,:,:) = PSRC(IE,:,:) * (0.5+SIGN(0.5,PRUCT(IE,:,:))) + PSRC(IE+1,:,:) * (0.5-SIGN(0.5,PRUCT(IE,:,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(IE,:,:) = 0.5 * (3.0*PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZFPOS2(IE,:,:) = 0.5 * (PSRC(IE,    :,:) + PSRC(IE+1,:,:))
    ZBPOS1(IE,:,:) = (PSRC(IE,:,:) - PSRC(IE-1,:,:))**2
    ZBPOS2(IE,:,:) = (PSRC(IE+1,:,:) - PSRC(IE,:,:))**2
!
    ZFNEG1(IE,:,:) = 0.5 * (3.0*PSRC(IE+1,:,:) - TEAST(:,:))
    ZFNEG2(IE,:,:) = 0.5 * (PSRC(IE,:,:) + PSRC(IE+1,:,:))
    ZBNEG1(IE,:,:) = (PSRC(IE+1,:,:) - TEAST(:,:))**2
    ZBNEG2(IE,:,:) = (PSRC(IE,  :,:) - PSRC(IE+1,:,:))**2
!
    ZOMP1(IE,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IE,:,:))**2
    ZOMP2(IE,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IE,:,:))**2
    ZOMN1(IE,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IE,:,:))**2
    ZOMN2(IE,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IE,:,:))**2
#else
    ZFPOS1(IE,:,:) = 0.5 * (3.0*PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZFPOS2(IE,:,:) = 0.5 * (PSRC(IE,    :,:) + PSRC(IE+1,:,:))
    ZBPOS1(IE,:,:) = BR_P2(PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZBPOS2(IE,:,:) = BR_P2(PSRC(IE+1,:,:) - PSRC(IE,:,:))
!
    ZFNEG1(IE,:,:) = 0.5 * (3.0*PSRC(IE+1,:,:) - TEAST(:,:))
    ZFNEG2(IE,:,:) = 0.5 * (PSRC(IE,:,:) + PSRC(IE+1,:,:))
    ZBNEG1(IE,:,:) = BR_P2(PSRC(IE+1,:,:) - TEAST(:,:))
    ZBNEG2(IE,:,:) = BR_P2(PSRC(IE,  :,:) - PSRC(IE+1,:,:))
!
    ZOMP1(IE,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IE,:,:))
    ZOMP2(IE,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IE,:,:))
    ZOMN1(IE,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IE,:,:))
    ZOMN2(IE,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IE,:,:))
#endif
!
    PR(IE,:,:) = (ZOMN2(IE,:,:)/(ZOMN1(IE,:,:)+ZOMN2(IE,:,:)) * ZFNEG2(IE,:,:) +  &
       (ZOMN1(IE,:,:)/(ZOMN1(IE,:,:)+ZOMN2(IE,:,:)) * ZFNEG1(IE,:,:))) * (0.5-SIGN(0.5,PRUCT(IE,:,:))) + &
                 (ZOMP2(IE,:,:)/(ZOMP1(IE,:,:)+ZOMP2(IE,:,:)) * ZFPOS2(IE,:,:) +  &
       (ZOMP1(IE,:,:)/(ZOMP1(IE,:,:)+ZOMP2(IE,:,:)) * ZFPOS1(IE,:,:))) * (0.5+SIGN(0.5,PRUCT(IE,:,:)))
!
  ENDIF
!
!      USE A THIRD ORDER UPSTREAM WENO SCHEME ELSEWHERE 
!
#ifndef MNH_BITREP
  ZFPOS1(IW:IE-1,:,:) = 0.5 * (3.0*PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZFPOS2(IW:IE-1,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBPOS1(IW:IE-1,:,:) = (PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))**2
  ZBPOS2(IW:IE-1,:,:) = (PSRC(IW+1:IE,:,:) - PSRC(IW:IE-1,  :,:))**2
!
  ZFNEG1(IW:IE-1,:,:) = 0.5 * (3.0*PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZFNEG2(IW:IE-1,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBNEG1(IW:IE-1,:,:) = (PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))**2
  ZBNEG2(IW:IE-1,:,:)   = (PSRC(IW:IE-1,:,:) - PSRC(IW+1:IE,:,:))**2
!
  ZOMP1(IW:IE-1,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IW:IE-1,:,:))**2
  ZOMP2(IW:IE-1,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IW:IE-1,:,:))**2
  ZOMN1(IW:IE-1,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IW:IE-1,:,:))**2
  ZOMN2(IW:IE-1,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IW:IE-1,:,:))**2
#else
  ZFPOS1(IW:IE-1,:,:) = 0.5 * (3.0*PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZFPOS2(IW:IE-1,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBPOS1(IW:IE-1,:,:) = BR_P2(PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZBPOS2(IW:IE-1,:,:) = BR_P2(PSRC(IW+1:IE,:,:) - PSRC(IW:IE-1,  :,:))
!
  ZFNEG1(IW:IE-1,:,:) = 0.5 * (3.0*PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZFNEG2(IW:IE-1,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBNEG1(IW:IE-1,:,:) = BR_P2(PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZBNEG2(IW:IE-1,:,:)   = BR_P2(PSRC(IW:IE-1,:,:) - PSRC(IW+1:IE,:,:))
!
  ZOMP1(IW:IE-1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IW:IE-1,:,:))
  ZOMP2(IW:IE-1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IW:IE-1,:,:))
  ZOMN1(IW:IE-1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IW:IE-1,:,:))
  ZOMN2(IW:IE-1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IW:IE-1,:,:))
#endif
!
    PR(IW:IE-1,:,:) = (ZOMN2(IW:IE-1,:,:)/(ZOMN1(IW:IE-1,:,:)+ZOMN2(IW:IE-1,:,:)) * ZFNEG2(IW:IE-1,:,:) +   &
       (ZOMN1(IW:IE-1,:,:)/(ZOMN1(IW:IE-1,:,:)+ZOMN2(IW:IE-1,:,:)) * ZFNEG1(IW:IE-1,:,:))) * (0.5-SIGN(0.5,PRUCT(IW:IE-1,:,:))) + &
       (ZOMP2(IW:IE-1,:,:)/(ZOMP1(IW:IE-1,:,:)+ZOMP2(IW:IE-1,:,:)) * ZFPOS2(IW:IE-1,:,:) +   &
       (ZOMP1(IW:IE-1,:,:)/(ZOMP1(IW:IE-1,:,:)+ZOMP2(IW:IE-1,:,:)) * ZFPOS1(IW:IE-1,:,:))) * (0.5+SIGN(0.5,PRUCT(IW:IE-1,:,:)))
!
END SELECT
!
PR = PR * PRUCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"ADVEC_WENO_K_2_UX end:PR")
END IF

END SUBROUTINE ADVEC_WENO_K_2_UX
!
!------------------------------------------------------------------------------
!
!     ############################################################
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_MX(HLBCX,PSRC, PRUCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_MX(HLBCX,PSRC, PRUCT, PR, TWEST, TEAST,            &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     ############################################################
!!
!!**** Computes PRUCT * PWT (or PRUCT * PVT). Upstream fluxes of W (or V) 
!!     variables in X direction.  
!!     Input PWT is on W Grid 'ie' (i,j,k) based on WGRID reference
!!     Output PR is on mass Grid 'ie' (i-1/2,j,k) based on WGRID reference  
!!
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*                
!!
!!    MODIFICATIONS
!!    -------------
!!
!------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODI_GET_HALO
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use mode_mppdb
#ifdef _OPENACC
use mode_msg
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX  ! X direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRUCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRUCT)
!
! output source term
!
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
REAL, DIMENSION(:,:), POINTER :: TWEST, TEAST
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TWEST,TEAST
!$acc declare copyin(TWEST,TEAST)
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IIB,IJB    ! Begining useful area in x,y,z directions
INTEGER :: IIE,IJE    ! End useful area in x,y,z directions
INTEGER::  IW,IE   ! Coordinate of third order diffusion area
!
INTEGER:: ILUOUT,IRESP   ! for prints
!
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "ADVEC_WENO_K_2_MX beg:PSRC")
  CALL MPPDB_CHECK(PRUCT,"ADVEC_WENO_K_2_MX beg:PRUCT")
END IF
!
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
!
!-----------------------------------------------------------------------------
!
!*       0.4.   INITIALIZE THE FIELD 
!               ---------------------
!
#ifndef _OPENACC
IF (PRESENT(TPHALO2)) THEN
  TWEST => TPHALO2%WEST
  TEAST => TPHALO2%EAST
END IF
#endif
!
!$acc kernels
PR(:,:,:) = 0.0
!
ZFPOS1 = 0.0
ZFPOS2 = 0.0
ZFNEG1 = 0.0
ZFNEG2 = 0.0 
ZBPOS1 = 0.0
ZBPOS2 = 0.0
ZBNEG1 = 0.0
ZBNEG2 = 0.0
ZOMP1  = 0.0
ZOMP2  = 0.0
ZOMN1  = 0.0
ZOMN2  = 0.0 
!
!------------------------------------------------------------------------------
!
SELECT CASE ( HLBCX(1) ) ! X direction LBC type: (1) for left side
!
!*       1.1    CYCLIC CASE IN THE X DIRECTION:
!
CASE ('CYCL')          ! In that case one must have HLBCX(1) == HLBCX(2)
#ifdef _OPENACC
  call Print_msg( NVERB_WARNING, 'GEN', 'ADVEC_WENO_K_2_MX', 'OpenACC: CYCL not yet tested' )
#endif
!
  IW=IIB
  IE=IIE
!
! intermediate fluxes for positive wind case
!
  ZFPOS1(IW+1:IE+1,:,:) = 0.5 * (3.0*PSRC(IW:IE,:,:) - PSRC(IW-1:IE-1,:,:))
  ZFPOS1(IW,       :,:) = 0.5 * (3.0*PSRC(IW-1, :,:) - TWEST(:,:))
!!  ZFPOS1(IW-1,     :,:) = - 999.
!
  ZFPOS2(IW:IE+1,:,:) = 0.5 * (PSRC(IW-1:IE,:,:) + PSRC(IW:IE+1,:,:))
  ZFPOS2(IW-1,   :,:) = 0.5 * (TWEST(:,:)        + PSRC(IW-1,   :,:))
!
! intermediate flux for negative wind case
!
  ZFNEG1(IW-1:IE,:,:) = 0.5 * (3.0*PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))
  ZFNEG1(IE+1,   :,:) = 0.5 * (3.0*PSRC(IE+1,   :,:) - TEAST(:,:))
!
  ZFNEG2(IW:IE+1,:,:) = 0.5 * (PSRC(IW:IE+1,:,:) + PSRC(IW-1:IE,:,:))
  ZFNEG2(IW-1,       :,:) = 0.5 * (PSRC(IW-1, :,:) + TWEST(:,:))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
  ZBPOS1(IW+1:IE+1,:,:) = (PSRC(IW:IE,:,:) - PSRC(IW-1:IE-1,:,:))**2
  ZBPOS1(IW,       :,:) = (PSRC(IW-1, :,:) - TWEST(:,:))**2
!!  ZBPOS1(IW-1,     :,:) = - 999.
!
  ZBPOS2(IW:IE+1,:,:) = (PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))**2
  ZBPOS2(IW-1,   :,:) = (PSRC(IW-1,   :,:) - TWEST(:,:))**2
!
! smoothness indicators for negative wind case
!
  ZBNEG1(IW-1:IE,:,:) = (PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))**2
  ZBNEG1(IE+1,   :,:) = (PSRC(IE+1,   :,:) - TEAST(:,:))**2
!
  ZBNEG2(IW:IE+1,:,:) = (PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))**2
  ZBNEG2(IW-1,   :,:) = (TWEST(:,:)        - PSRC(IW-1,:,:))**2
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / (ZEPS + ZBPOS1)**2
  ZOMP2 = ZGAMMA2 / (ZEPS + ZBPOS2)**2
  ZOMN1 = ZGAMMA1 / (ZEPS + ZBNEG1)**2
  ZOMN2 = ZGAMMA2 / (ZEPS + ZBNEG2)**2
#else
!
! smoothness indicators for positive wind case
!
  ZBPOS1(IW+1:IE+1,:,:) = BR_P2(PSRC(IW:IE,:,:) - PSRC(IW-1:IE-1,:,:))
  ZBPOS1(IW,       :,:) = BR_P2(PSRC(IW-1, :,:) - TWEST(:,:))
!!  ZBPOS1(IW-1,     :,:) = - 999.
!
  ZBPOS2(IW:IE+1,:,:) = BR_P2(PSRC(IW:IE+1,:,:) - PSRC(IW-1:IE,:,:))
  ZBPOS2(IW-1,   :,:) = BR_P2(PSRC(IW-1,   :,:) - TWEST(:,:))
!
! smoothness indicators for negative wind case
!
  ZBNEG1(IW-1:IE,:,:) = BR_P2(PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))
  ZBNEG1(IE+1,   :,:) = BR_P2(PSRC(IE+1,   :,:) - TEAST(:,:))
!
  ZBNEG2(IW:IE+1,:,:) = BR_P2(PSRC(IW-1:IE,:,:) - PSRC(IW:IE+1,:,:))
  ZBNEG2(IW-1,   :,:) = BR_P2(TWEST(:,:)        - PSRC(IW-1,:,:))
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1)
  ZOMP2 = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2)
  ZOMN1 = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1)
  ZOMN2 = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2)
#endif
!
! WENO fluxes
!
  PR = (ZOMP2/(ZOMP1+ZOMP2) * ZFPOS2 +                            &
       (ZOMP1/(ZOMP1+ZOMP2) * ZFPOS1)) * (0.5+SIGN(0.5,PRUCT )) + &
       (ZOMN2/(ZOMN1+ZOMN2) * ZFNEG2 +                            &
       (ZOMN1/(ZOMN1+ZOMN2) * ZFNEG1)) * (0.5-SIGN(0.5,PRUCT ))
!
!
!       OPEN, WALL, NEST CASE IN THE X DIRECTION
!
CASE ('OPEN','WALL','NEST')
!
  IW=IIB
  IE=IIE
!
!       USE A FIRST ORDER UPSTREAM SCHEME AT THE PHYSICAL BORDER
!
  IF(LWEST_ll()) THEN
    PR(IW,:,:) = PSRC(IW-1,:,:) * (0.5+SIGN(0.5,PRUCT(IW,:,:))) + PSRC(IW,:,:) * (0.5-SIGN(0.5,PRUCT(IW,:,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(IW,:,:) = 0.5 * (3.0*PSRC(IW-1, :,:) - TWEST(:,:))
    ZFPOS2(IW,:,:) = 0.5 * (PSRC(IW-1,     :,:) + PSRC(IW,     :,:))
    ZBPOS1(IW,:,:) = (PSRC(IW-1,:,:) - TWEST(:,:))**2
    ZBPOS2(IW,:,:) = (PSRC(IW,  :,:) - PSRC(IW-1,:,:))**2
!
    ZFNEG1(IW,:,:) = 0.5 * (3.0*PSRC(IW,:,:) - PSRC(IW+1,:,:))
    ZFNEG2(IW,:,:) = 0.5 * (PSRC(IW,    :,:) + PSRC(IW-1,:,:))
    ZBNEG1(IW,:,:) = (PSRC(IW,:,:) - PSRC(IW+1,:,:))**2
    ZBNEG2(IW,:,:) = (PSRC(IW-1,:,:) - PSRC(IW,:,:))**2
!
    ZOMP1(IW,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IW,:,:))**2
    ZOMP2(IW,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IW,:,:))**2
    ZOMN1(IW,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IW,:,:))**2
    ZOMN2(IW,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IW,:,:))**2
#else
    ZFPOS1(IW,:,:) = 0.5 * (3.0*PSRC(IW-1, :,:) - TWEST(:,:))
    ZFPOS2(IW,:,:) = 0.5 * (PSRC(IW-1,     :,:) + PSRC(IW,     :,:))
    ZBPOS1(IW,:,:) = BR_P2(PSRC(IW-1,:,:) - TWEST(:,:))
    ZBPOS2(IW,:,:) = BR_P2(PSRC(IW,  :,:) - PSRC(IW-1,:,:))
!
    ZFNEG1(IW,:,:) = 0.5 * (3.0*PSRC(IW,:,:) - PSRC(IW+1,:,:))
    ZFNEG2(IW,:,:) = 0.5 * (PSRC(IW,    :,:) + PSRC(IW-1,:,:))
    ZBNEG1(IW,:,:) = BR_P2(PSRC(IW,:,:) - PSRC(IW+1,:,:))
    ZBNEG2(IW,:,:) = BR_P2(PSRC(IW-1,:,:) - PSRC(IW,:,:))
!
    ZOMP1(IW,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IW,:,:))
    ZOMP2(IW,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IW,:,:))
    ZOMN1(IW,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IW,:,:))
    ZOMN2(IW,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IW,:,:))
#endif
!
    PR(IW,:,:) = (ZOMP2(IW,:,:)/(ZOMP1(IW,:,:)+ZOMP2(IW,:,:)) * ZFPOS2(IW,:,:) +  &
       (ZOMP1(IW,:,:)/(ZOMP1(IW,:,:)+ZOMP2(IW,:,:)) * ZFPOS1(IW,:,:))) * (0.5+SIGN(0.5,PRUCT(IW,:,:))) + &
       (ZOMN2(IW,:,:)/(ZOMN1(IW,:,:)+ZOMN2(IW,:,:)) * ZFNEG2(IW,:,:) +   &
       (ZOMN1(IW,:,:)/(ZOMN1(IW,:,:)+ZOMN2(IW,:,:)) * ZFNEG1(IW,:,:))) * (0.5-SIGN(0.5,PRUCT(IW,:,:)))
!
  ENDIF
!
  IF(LEAST_ll()) THEN
    PR(IE+1,:,:) = PSRC(IE,:,:) * (0.5+SIGN(0.5,PRUCT(IE+1,:,:))) + PSRC(IE+1,:,:) * (0.5-SIGN(0.5,PRUCT(IE+1,:,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(IE+1,:,:) = 0.5 * (3.0*PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZFPOS2(IE+1,:,:) = 0.5 * (PSRC(IE,    :,:) + PSRC(IE+1,:,:))
    ZBPOS1(IE+1,:,:) = (PSRC(IE,:,:) - PSRC(IE-1,:,:))**2
    ZBPOS2(IE+1,:,:) = (PSRC(IE+1,:,:) - PSRC(IE,:,:))**2
!
    ZFNEG1(IE+1,:,:) = 0.5 * (3.0*PSRC(IE+1,:,:) - TEAST(:,:))
    ZFNEG2(IE+1,:,:) = 0.5 * (PSRC(IE+1,    :,:) + PSRC(IE,:,:))
    ZBNEG1(IE+1,:,:) = (PSRC(IE+1,:,:) - TEAST(:,:))**2
    ZBNEG2(IE+1,:,:) = (PSRC(IE,  :,:) - PSRC(IE+1,:,:))**2
!
    ZOMP1(IE+1,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IE+1,:,:))**2
    ZOMP2(IE+1,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IE+1,:,:))**2
    ZOMN1(IE+1,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IE+1,:,:))**2
    ZOMN2(IE+1,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IE+1,:,:))**2
#else
    ZFPOS1(IE+1,:,:) = 0.5 * (3.0*PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZFPOS2(IE+1,:,:) = 0.5 * (PSRC(IE,    :,:) + PSRC(IE+1,:,:))
    ZBPOS1(IE+1,:,:) = BR_P2(PSRC(IE,:,:) - PSRC(IE-1,:,:))
    ZBPOS2(IE+1,:,:) = BR_P2(PSRC(IE+1,:,:) - PSRC(IE,:,:))
!
    ZFNEG1(IE+1,:,:) = 0.5 * (3.0*PSRC(IE+1,:,:) - TEAST(:,:))
    ZFNEG2(IE+1,:,:) = 0.5 * (PSRC(IE+1,    :,:) + PSRC(IE,:,:))
    ZBNEG1(IE+1,:,:) = BR_P2(PSRC(IE+1,:,:) - TEAST(:,:))
    ZBNEG2(IE+1,:,:) = BR_P2(PSRC(IE,  :,:) - PSRC(IE+1,:,:))
!
    ZOMP1(IE+1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IE+1,:,:))
    ZOMP2(IE+1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IE+1,:,:))
    ZOMN1(IE+1,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IE+1,:,:))
    ZOMN2(IE+1,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IE+1,:,:))
#endif
!
    PR(IE+1,:,:) = (ZOMP2(IE+1,:,:)/(ZOMP1(IE+1,:,:)+ZOMP2(IE+1,:,:)) * ZFPOS2(IE+1,:,:) +  &
       (ZOMP1(IE+1,:,:)/(ZOMP1(IE+1,:,:)+ZOMP2(IE+1,:,:)) * ZFPOS1(IE+1,:,:))) * (0.5+SIGN(0.5,PRUCT(IE+1,:,:))) + &
       (ZOMN2(IE+1,:,:)/(ZOMN1(IE+1,:,:)+ZOMN2(IE+1,:,:)) * ZFNEG2(IE+1,:,:) +   &
       (ZOMN1(IE+1,:,:)/(ZOMN1(IE+1,:,:)+ZOMN2(IE+1,:,:)) * ZFNEG1(IE+1,:,:))) * (0.5-SIGN(0.5,PRUCT(IE+1,:,:)))
!
  ENDIF
!
!      USE A THIRD ORDER UPSTREAM WENO SCHEME ELSEWHERE 
!
#ifndef MNH_BITREP
  ZFPOS1(IW+1:IE,:,:) = 0.5 * (3.0*PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZFPOS2(IW+1:IE,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBPOS1(IW+1:IE,:,:) = (PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))**2
  ZBPOS2(IW+1:IE,:,:) = (PSRC(IW+1:IE,:,:) - PSRC(IW:IE-1,:,:))**2
!
  ZFNEG1(IW+1:IE,:,:) = 0.5 * (3.0*PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZFNEG2(IW+1:IE,:,:) = 0.5 * (PSRC(IW+1:IE,    :,:) + PSRC(IW:IE-1,  :,:))
  ZBNEG1(IW+1:IE,:,:) = (PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))**2
  ZBNEG2(IW+1:IE,:,:) = (PSRC(IW:IE-1,:,:) - PSRC(IW+1:IE,:,:))**2
!
  ZOMP1(IW+1:IE,:,:) = ZGAMMA1 / (ZEPS + ZBPOS1(IW+1:IE,:,:))**2
  ZOMP2(IW+1:IE,:,:) = ZGAMMA2 / (ZEPS + ZBPOS2(IW+1:IE,:,:))**2
  ZOMN1(IW+1:IE,:,:) = ZGAMMA1 / (ZEPS + ZBNEG1(IW+1:IE,:,:))**2
  ZOMN2(IW+1:IE,:,:) = ZGAMMA2 / (ZEPS + ZBNEG2(IW+1:IE,:,:))**2
#else
  ZFPOS1(IW+1:IE,:,:) = 0.5 * (3.0*PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZFPOS2(IW+1:IE,:,:) = 0.5 * (PSRC(IW:IE-1,    :,:) + PSRC(IW+1:IE,  :,:))
  ZBPOS1(IW+1:IE,:,:) = BR_P2(PSRC(IW:IE-1,:,:) - PSRC(IW-1:IE-2,:,:))
  ZBPOS2(IW+1:IE,:,:) = BR_P2(PSRC(IW+1:IE,:,:) - PSRC(IW:IE-1,:,:))
!
  ZFNEG1(IW+1:IE,:,:) = 0.5 * (3.0*PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZFNEG2(IW+1:IE,:,:) = 0.5 * (PSRC(IW+1:IE,    :,:) + PSRC(IW:IE-1,  :,:))
  ZBNEG1(IW+1:IE,:,:) = BR_P2(PSRC(IW+1:IE,:,:) - PSRC(IW+2:IE+1,:,:))
  ZBNEG2(IW+1:IE,:,:) = BR_P2(PSRC(IW:IE-1,:,:) - PSRC(IW+1:IE,:,:))
!
  ZOMP1(IW+1:IE,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(IW+1:IE,:,:))
  ZOMP2(IW+1:IE,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(IW+1:IE,:,:))
  ZOMN1(IW+1:IE,:,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(IW+1:IE,:,:))
  ZOMN2(IW+1:IE,:,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(IW+1:IE,:,:))
#endif
!
  PR(IW+1:IE,:,:) = (ZOMP2(IW+1:IE,:,:)/(ZOMP1(IW+1:IE,:,:)+ZOMP2(IW+1:IE,:,:)) * ZFPOS2(IW+1:IE,:,:) + &
       (ZOMP1(IW+1:IE,:,:)/(ZOMP1(IW+1:IE,:,:)+ZOMP2(IW+1:IE,:,:)) * ZFPOS1(IW+1:IE,:,:))) * (0.5+SIGN(0.5,PRUCT(IW+1:IE,:,:))) + &
       (ZOMN2(IW+1:IE,:,:)/(ZOMN1(IW+1:IE,:,:)+ZOMN2(IW+1:IE,:,:)) * ZFNEG2(IW+1:IE,:,:) +              &
       (ZOMN1(IW+1:IE,:,:)/(ZOMN1(IW+1:IE,:,:)+ZOMN2(IW+1:IE,:,:)) * ZFNEG1(IW+1:IE,:,:))) * (0.5-SIGN(0.5,PRUCT(IW+1:IE,:,:)))
!
END SELECT
!
PR = PR * PRUCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"ADVEC_WENO_K_2_MX end:PR")
END IF

END SUBROUTINE ADVEC_WENO_K_2_MX
!
!-------------------------------------------------------------------------------
!
!     ############################################################
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_MY(HLBCY,PSRC, PRVCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_MY(HLBCY,PSRC, PRVCT, PR, TNORTH, TSOUTH,          &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     ############################################################
!!
!!****  Computes PRVCT * PUT (or PRVCT * PWT). Upstream fluxes of U (or W) 
!!      variables in Y direction.  
!!      Input PUT is on U Grid 'ie' (i,j,k) based on UGRID reference
!!      Output PR is on mass Grid 'ie' (i,j-1/2,k) based on UGRID reference 
!!
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*                 
!!
!!    MODIFICATIONS
!!    -------------
!!
!-------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODI_GET_HALO
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use mode_mppdb
#ifdef _OPENACC
use mode_msg
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY  ! Y direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRVCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRVCT)
!
! output source term
!
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
REAL, DIMENSION(:,:), POINTER :: TNORTH, TSOUTH
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TNORTH, TSOUTH
!$acc declare copyin(TNORTH,TSOUTH)
#endif
!
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IIB,IJB    ! Begining useful area in x,y,z directions
INTEGER :: IIE,IJE    ! End useful area in x,y,z directions
INTEGER::  IS,IN      ! Coordinate of third order diffusion area
!
INTEGER:: ILUOUT,IRESP   ! for prints
!
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!-----------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "ADVEC_WENO_K_2_MY beg:PSRC")
  CALL MPPDB_CHECK(PRVCT,"ADVEC_WENO_K_2_MY beg:PRVCT")
END IF
!
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
!
!---------------------------------------------------------------------------
!
!*       0.4.   INITIALIZE THE FIELD 
!               ---------------------
!
#ifndef _OPENACC
IF (PRESENT(TPHALO2)) THEN
  TNORTH => TPHALO2%NORTH
  TSOUTH => TPHALO2%SOUTH
END IF
#endif
!
!$acc kernels
PR(:,:,:) = 0.0
!
ZFPOS1 = 0.0
ZFPOS2 = 0.0
ZFNEG1 = 0.0
ZFNEG2 = 0.0
ZBPOS1 = 0.0
ZBPOS2 = 0.0
ZBNEG1 = 0.0
ZBNEG2 = 0.0
ZOMP1  = 0.0
ZOMP2  = 0.0
ZOMN1  = 0.0
ZOMN2  = 0.0 
!
!-------------------------------------------------------------------------------
!
SELECT CASE ( HLBCY(1) ) ! 
!
!*       1.1    CYCLIC CASE IN THE Y DIRECTION:
!
CASE ('CYCL')          ! In that case one must have HLBCY(1) == HLBCY(2)
#ifdef _OPENACC
  call Print_msg( NVERB_WARNING, 'GEN', 'ADVEC_WENO_K_2_MY', 'OpenACC: CYCL not yet tested' )
#endif
!
  IS=IJB
  IN=IJE
!
! intermediate fluxes for positive wind case
!
  ZFPOS1(:,IS+1:IN+1,:) = 0.5 * (3.0*PSRC(:,IS:IN,:) - PSRC(:,IS-1:IN-1,:))
  ZFPOS1(:,IS,       :) = 0.5 * (3.0*PSRC(:,IS-1, :) - TSOUTH(:,:))
!!  ZFPOS1(:,IS-1,     :) = - 999.
!
  ZFPOS2(:,IS:IN+1,:) = 0.5 * (PSRC(:,IS-1:IN,:) + PSRC(:,IS:IN+1,:))
  ZFPOS2(:,IS-1,   :) = 0.5 * (TSOUTH(:,:)       + PSRC(:,IS-1,   :))
!
  ZFNEG1(:,IS-1:IN,:) = 0.5 * (3.0*PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))
  ZFNEG1(:,IN+1,   :) = 0.5 * (3.0*PSRC(:,IN+1,   :) - TNORTH(:,:))
!
  ZFNEG2(:,IS:IN+1,:) = 0.5 * (PSRC(:,IS:IN+1,:) + PSRC(:,IS-1:IN,:))
  ZFNEG2(:,IS-1,   :) = 0.5 * (PSRC(:,IS-1,   :) + TSOUTH(:,:))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
  ZBPOS1(:,IS+1:IN+1,:) = (PSRC(:,IS:IN,:) - PSRC(:,IS-1:IN-1,:))**2
  ZBPOS1(:,IS,       :) = (PSRC(:,IS-1,   :) - TSOUTH(:,:))**2
!!  ZBPOS1(:,IS-1,     :) = - 999.
!
  ZBPOS2(:,IS:IN+1,:) = (PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))**2
  ZBPOS2(:,IS-1,   :) = (PSRC(:,IS-1,   :) - TSOUTH(:,:))**2
!
! smoothness indicators for negative wind case
!
  ZBNEG1(:,IS-1:IN,:) = (PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))**2
  ZBNEG1(:,IN+1,   :) = (PSRC(:,IN+1,   :) - TNORTH(:,:))**2
!
  ZBNEG2(:,IS:IN+1,:) = (PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))**2
  ZBNEG2(:,IS-1,   :) = (TSOUTH(:,:)       - PSRC(:,IS-1,:))**2
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / (ZEPS + ZBPOS1)**2
  ZOMP2 = ZGAMMA2 / (ZEPS + ZBPOS2)**2
  ZOMN1 = ZGAMMA1 / (ZEPS + ZBNEG1)**2
  ZOMN2 = ZGAMMA2 / (ZEPS + ZBNEG2)**2
#else
!
! smoothness indicators for positive wind case
!
  ZBPOS1(:,IS+1:IN+1,:) = BR_P2(PSRC(:,IS:IN,:) - PSRC(:,IS-1:IN-1,:))
  ZBPOS1(:,IS,       :) = BR_P2(PSRC(:,IS-1,   :) - TSOUTH(:,:))
!!  ZBPOS1(:,IS-1,     :) = - 999.
!
  ZBPOS2(:,IS:IN+1,:) = BR_P2(PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))
  ZBPOS2(:,IS-1,   :) = BR_P2(PSRC(:,IS-1,   :) - TSOUTH(:,:))
!
! smoothness indicators for negative wind case
!
  ZBNEG1(:,IS-1:IN,:) = BR_P2(PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))
  ZBNEG1(:,IN+1,   :) = BR_P2(PSRC(:,IN+1,   :) - TNORTH(:,:))
!
  ZBNEG2(:,IS:IN+1,:) = BR_P2(PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))
  ZBNEG2(:,IS-1,   :) = BR_P2(TSOUTH(:,:)       - PSRC(:,IS-1,:))
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1)
  ZOMP2 = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2)
  ZOMN1 = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1)
  ZOMN2 = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2)
#endif
!
! WENO fluxes
!
  PR = (ZOMP2/(ZOMP1+ZOMP2) * ZFPOS2 +                           &
       (ZOMP1/(ZOMP1+ZOMP2) * ZFPOS1)) * (0.5+SIGN(0.5,PRVCT)) + &
       (ZOMN2/(ZOMN1+ZOMN2) * ZFNEG2 +                           &
       (ZOMN1/(ZOMN1+ZOMN2) * ZFNEG1)) * (0.5-SIGN(0.5,PRVCT))
!
!
!       OPEN, WALL, NEST CASE IN THE Y DIRECTION
!
CASE ('OPEN','WALL','NEST')
!
  IS=IJB
  IN=IJE
!
!       USE A FIRST ORDER UPSTREAM SCHEME AT THE PHYSICAL BORDER
!
  IF(LSOUTH_ll()) THEN
    PR(:,IS,:) = PSRC(:,IS-1,:) * (0.5+SIGN(0.5,PRVCT(:,IS,:))) + PSRC(:,IS,:) * (0.5-SIGN(0.5,PRVCT(:,IS,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(:,IS,:) = 0.5 * (3.0*PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZFPOS2(:,IS,:) = 0.5 * (PSRC(:,IS-1,:) + PSRC(:,IS,:))
    ZBPOS1(:,IS,:) = (PSRC(:,IS-1,:) - TSOUTH(:,:))**2
    ZBPOS2(:,IS,:) = (PSRC(:,IS,  :) - PSRC(:,IS-1,:))**2
!
    ZFNEG1(:,IS,:) = 0.5 * (3.0*PSRC(:,IS,:) - PSRC(:,IS+1,:))
    ZFNEG2(:,IS,:) = 0.5 * (PSRC(:,IS,    :) + PSRC(:,IS-1,:))
    ZBNEG1(:,IS,:) = (PSRC(:,IS,  :) - PSRC(:,IS+1,:))**2
    ZBNEG2(:,IS,:) = (PSRC(:,IS-1,:) - PSRC(:,IS,  :))**2
!
    ZOMP1(:,IS,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IS,:))**2
    ZOMP2(:,IS,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IS,:))**2
    ZOMN1(:,IS,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IS,:))**2
    ZOMN2(:,IS,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IS,:))**2
#else
    ZFPOS1(:,IS,:) = 0.5 * (3.0*PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZFPOS2(:,IS,:) = 0.5 * (PSRC(:,IS-1,:) + PSRC(:,IS,:))
    ZBPOS1(:,IS,:) = BR_P2(PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZBPOS2(:,IS,:) = BR_P2(PSRC(:,IS,  :) - PSRC(:,IS-1,:))
!
    ZFNEG1(:,IS,:) = 0.5 * (3.0*PSRC(:,IS,:) - PSRC(:,IS+1,:))
    ZFNEG2(:,IS,:) = 0.5 * (PSRC(:,IS,    :) + PSRC(:,IS-1,:))
    ZBNEG1(:,IS,:) = BR_P2(PSRC(:,IS,  :) - PSRC(:,IS+1,:))
    ZBNEG2(:,IS,:) = BR_P2(PSRC(:,IS-1,:) - PSRC(:,IS,  :))
!
    ZOMP1(:,IS,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IS,:))
    ZOMP2(:,IS,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IS,:))
    ZOMN1(:,IS,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IS,:))
    ZOMN2(:,IS,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IS,:))
#endif
!
    PR(:,IS,:) = (ZOMP2(:,IS,:)/(ZOMP1(:,IS,:)+ZOMP2(:,IS,:)) * ZFPOS2(:,IS,:) + &
       (ZOMP1(:,IS,:)/(ZOMP1(:,IS,:)+ZOMP2(:,IS,:)) * ZFPOS1(:,IS,:))) * (0.5+SIGN(0.5,PRVCT(:,IS,:))) + &
       (ZOMN2(:,IS,:)/(ZOMN1(:,IS,:)+ZOMN2(:,IS,:)) * ZFNEG2(:,IS,:) +  &
       (ZOMN1(:,IS,:)/(ZOMN1(:,IS,:)+ZOMN2(:,IS,:)) * ZFNEG1(:,IS,:))) * (0.5-SIGN(0.5,PRVCT(:,IS,:)))
!
  ENDIF
!
  IF(LNORTH_ll()) THEN
    PR(:,IN+1,:) = PSRC(:,IN,:) * (0.5+SIGN(0.5,PRVCT(:,IN+1,:))) + PSRC(:,IN+1,:) * (0.5-SIGN(0.5,PRVCT(:,IN+1,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(:,IN+1,:) = 0.5 * (3.0*PSRC(:,IN,:) - PSRC(:,IN-1,:))
    ZFPOS2(:,IN+1,:) = 0.5 * (PSRC(:,IN,    :) + PSRC(:,IN+1,:))
    ZBPOS1(:,IN+1,:) = (PSRC(:,IN,:) - PSRC(:,IN-1,:))**2
    ZBPOS2(:,IN+1,:) = (PSRC(:,IN+1,:) - PSRC(:,IN,:))**2
!
    ZFNEG1(:,IN+1,:) = 0.5 * (3.0*PSRC(:,IN+1,:) - TNORTH(:,:))
    ZFNEG2(:,IN+1,:) = 0.5 * (PSRC(:,IN+1,    :) + PSRC(:,IN,:))
    ZBNEG1(:,IN+1,:) = (PSRC(:,IN+1,:) - TNORTH(:,:))**2
    ZBNEG2(:,IN+1,:) = (PSRC(:,IN,  :) - PSRC(:,IN+1,:))**2
!
    ZOMP1(:,IN+1,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IN+1,:))**2
    ZOMP2(:,IN+1,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IN+1,:))**2
    ZOMN1(:,IN+1,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IN+1,:))**2
    ZOMN2(:,IN+1,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IN+1,:))**2
#else
    ZFPOS1(:,IN+1,:) = 0.5 * (3.0*PSRC(:,IN,:) - PSRC(:,IN-1,:))
    ZFPOS2(:,IN+1,:) = 0.5 * (PSRC(:,IN,    :) + PSRC(:,IN+1,:))
    ZBPOS1(:,IN+1,:) = BR_P2(PSRC(:,IN,:) - PSRC(:,IN-1,:))
    ZBPOS2(:,IN+1,:) = BR_P2(PSRC(:,IN+1,:) - PSRC(:,IN,:))
!
    ZFNEG1(:,IN+1,:) = 0.5 * (3.0*PSRC(:,IN+1,:) - TNORTH(:,:))
    ZFNEG2(:,IN+1,:) = 0.5 * (PSRC(:,IN+1,    :) + PSRC(:,IN,:))
    ZBNEG1(:,IN+1,:) = BR_P2(PSRC(:,IN+1,:) - TNORTH(:,:))
    ZBNEG2(:,IN+1,:) = BR_P2(PSRC(:,IN,  :) - PSRC(:,IN+1,:))
!
    ZOMP1(:,IN+1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IN+1,:))
    ZOMP2(:,IN+1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IN+1,:))
    ZOMN1(:,IN+1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IN+1,:))
    ZOMN2(:,IN+1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IN+1,:))
#endif
!
    PR(:,IN+1,:) = (ZOMP2(:,IN+1,:)/(ZOMP1(:,IN+1,:)+ZOMP2(:,IN+1,:)) * ZFPOS2(:,IN+1,:) +   &
       (ZOMP1(:,IN+1,:)/(ZOMP1(:,IN+1,:)+ZOMP2(:,IN+1,:)) * ZFPOS1(:,IN+1,:))) * (0.5+SIGN(0.5,PRVCT(:,IN+1,:))) + &
       (ZOMN2(:,IN+1,:)/(ZOMN1(:,IN+1,:)+ZOMN2(:,IN+1,:)) * ZFNEG2(:,IN+1,:) +     &
       (ZOMN1(:,IN+1,:)/(ZOMN1(:,IN+1,:)+ZOMN2(:,IN+1,:)) * ZFNEG1(:,IN+1,:))) * (0.5-SIGN(0.5,PRVCT(:,IN+1,:)))
!
  ENDIF
!
!      USE A THIRD ORDER UPSTREAM WENO SCHEME ELSEWHERE 
!
#ifndef MNH_BITREP
  ZFPOS1(:,IS+1:IN,:) = 0.5 * (3.0*PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZFPOS2(:,IS+1:IN,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBPOS1(:,IS+1:IN,:) = (PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))**2
  ZBPOS2(:,IS+1:IN,:) = (PSRC(:,IS+1:IN,:) - PSRC(:,IS:IN-1,  :))**2
!
  ZFNEG1(:,IS+1:IN,:) = 0.5 * (3.0*PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZFNEG2(:,IS+1:IN,:) = 0.5 * (PSRC(:,IS+1:IN,    :) + PSRC(:,IS:IN-1,  :))
  ZBNEG1(:,IS+1:IN,:) = (PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))**2
  ZBNEG2(:,IS+1:IN,:) = (PSRC(:,IS:IN-1,:) - PSRC(:,IS+1:IN,:))**2
!
  ZOMP1(:,IS+1:IN,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IS+1:IN,:))**2
  ZOMP2(:,IS+1:IN,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IS+1:IN,:))**2
  ZOMN1(:,IS+1:IN,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IS+1:IN,:))**2
  ZOMN2(:,IS+1:IN,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IS+1:IN,:))**2
#else
  ZFPOS1(:,IS+1:IN,:) = 0.5 * (3.0*PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZFPOS2(:,IS+1:IN,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBPOS1(:,IS+1:IN,:) = BR_P2(PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZBPOS2(:,IS+1:IN,:) = BR_P2(PSRC(:,IS+1:IN,:) - PSRC(:,IS:IN-1,  :))
!
  ZFNEG1(:,IS+1:IN,:) = 0.5 * (3.0*PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZFNEG2(:,IS+1:IN,:) = 0.5 * (PSRC(:,IS+1:IN,    :) + PSRC(:,IS:IN-1,  :))
  ZBNEG1(:,IS+1:IN,:) = BR_P2(PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZBNEG2(:,IS+1:IN,:) = BR_P2(PSRC(:,IS:IN-1,:) - PSRC(:,IS+1:IN,:))
!
  ZOMP1(:,IS+1:IN,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IS+1:IN,:))
  ZOMP2(:,IS+1:IN,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IS+1:IN,:))
  ZOMN1(:,IS+1:IN,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IS+1:IN,:))
  ZOMN2(:,IS+1:IN,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IS+1:IN,:))
#endif
!
  PR(:,IS+1:IN,:) = (ZOMP2(:,IS+1:IN,:)/(ZOMP1(:,IS+1:IN,:)+ZOMP2(:,IS+1:IN,:)) * ZFPOS2(:,IS+1:IN,:) +  &
       (ZOMP1(:,IS+1:IN,:)/(ZOMP1(:,IS+1:IN,:)+ZOMP2(:,IS+1:IN,:)) * ZFPOS1(:,IS+1:IN,:))) * (0.5+SIGN(0.5,PRVCT(:,IS+1:IN,:))) + &
       (ZOMN2(:,IS+1:IN,:)/(ZOMN1(:,IS+1:IN,:)+ZOMN2(:,IS+1:IN,:)) * ZFNEG2(:,IS+1:IN,:) + &
       (ZOMN1(:,IS+1:IN,:)/(ZOMN1(:,IS+1:IN,:)+ZOMN2(:,IS+1:IN,:)) * ZFNEG1(:,IS+1:IN,:))) * (0.5-SIGN(0.5,PRVCT(:,IS+1:IN,:)))
!
END SELECT
!
PR = PR * PRVCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"ADVEC_WENO_K_2_MY end:PR")
END IF

END SUBROUTINE ADVEC_WENO_K_2_MY
!-------------------------------------------------------------------------------
!
!     #############################################################
#ifndef _OPENACC
      SUBROUTINE ADVEC_WENO_K_2_VY(HLBCY, PSRC, PRVCT, PR, TPHALO2)
#else
      SUBROUTINE ADVEC_WENO_K_2_VY(HLBCY, PSRC, PRVCT, PR, TNORTH, TSOUTH,         &
                                   ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                                   ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     #############################################################
!!
!!**** Computes PRVCT * PVT. Upstream fluxes of V in Y direction.  
!!     Input PVT is on V Grid 'ie' (i,j,k) based on VGRID reference
!!     Output PR is on mass Grid 'ie' (i,j+1/2,k) based on VGRID reference
!!
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*
!!
!!    MODIFICATIONS
!!    -------------
!!
!-------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_LUNIT
USE MODD_CONF
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODI_GET_HALO
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use mode_mppdb
#ifdef _OPENACC
use mode_msg
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY  ! Y direction LBC type
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on U grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRVCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRVCT)
!
! output source term
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#ifndef _OPENACC
TYPE(HALO2_ll), OPTIONAL, POINTER :: TPHALO2      ! halo2 for the field at t
REAL, DIMENSION(:,:), POINTER :: TNORTH, TSOUTH
#else
REAL, DIMENSION(:,:), INTENT(IN) :: TNORTH, TSOUTH
!$acc declare copyin(TNORTH,TSOUTH)
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IIB,IJB    ! Begining useful area in x,y,z directions
INTEGER :: IIE,IJE    ! End useful area in x,y,z directions
INTEGER::  IS,IN      ! Coordinate of third order diffusion area
!
INTEGER:: ILUOUT,IRESP   ! for prints
!
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!----------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "ADVEC_WENO_K_2_VY beg:PSRC")
  CALL MPPDB_CHECK(PRVCT,"ADVEC_WENO_K_2_VY beg:PRVCT")
END IF
!
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
!
!--------------------------------------------------------------------------
!
!*       0.4.   INITIALIZE THE FIELD 
!               ---------------------
!
#ifndef _OPENACC
IF (PRESENT(TPHALO2)) THEN
  TNORTH => TPHALO2%NORTH
  TSOUTH => TPHALO2%SOUTH
END IF
#endif
!
!$acc kernels
PR(:,:,:) = 0.0
!
ZFPOS1 = 0.0
ZFPOS2 = 0.0
ZFNEG1 = 0.0
ZFNEG2 = 0.0
ZBPOS1 = 0.0
ZBPOS2 = 0.0
ZBNEG1 = 0.0
ZBNEG2 = 0.0
ZOMP1  = 0.0
ZOMP2  = 0.0
ZOMN1  = 0.0
ZOMN2  = 0.0
!
!-------------------------------------------------------------------------------
!
SELECT CASE ( HLBCY(1) ) ! Y direction LBC type: (1) for left side
!
!*       1.1    CYCLIC CASE IN THE Y DIRECTION:
!
CASE ('CYCL')          ! In that case one must have HLBCX(1) == HLBCX(2)
#ifdef _OPENACC
  call Print_msg( NVERB_WARNING, 'GEN', 'ADVEC_WENO_K_2_VY', 'OpenACC: CYCL not yet tested' )
#endif
!
  IS=IJB
  IN=IJE
!
! intermediate fluxes for positive wind case
!
  ZFPOS1(:,IS:IN+1,:) = 0.5 * (3.0*PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))
  ZFPOS1(:,IS-1,   :) = 0.5 * (3.0*PSRC(:,IS-1,   :) - TSOUTH(:,:))
!
  ZFPOS2(:,IS-1:IN,:) = 0.5 * (PSRC(:,IS-1:IN,:) + PSRC(:,IS:IN+1,:))
  ZFPOS2(:,IN+1,   :) = 0.5 * (PSRC(:,IN+1,   :) + TNORTH(:,:))
!
! intermediate flux for negative wind case
!
  ZFNEG1(:,IS-1:IN-1,:) = 0.5 * (3.0*PSRC(:,IS:IN,:) - PSRC(:,IS+1:IN+1,:))
  ZFNEG1(:,IN,   :)     = 0.5 * (3.0*PSRC(:,IN+1,   :) - TNORTH(:,:))
!
  ZFNEG2(:,IS-1:IN,:) = 0.5 * (PSRC(:,IS-1:IN,:) + PSRC(:,IS:IN+1,:))
  ZFNEG2(:,IN+1,   :) = 0.5 * (PSRC(:,IN+1,   :) + TNORTH(:,:))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
  ZBPOS1(:,IS:IN+1,:) = (PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))**2
  ZBPOS1(:,IS-1,   :) = (PSRC(:,IS-1,   :) - TSOUTH(:,:))**2
!
  ZBPOS2(:,IS-1:IN,:) = (PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))**2
  ZBPOS2(:,IN+1,   :) = (TNORTH(:,:)       - PSRC(:,IN+1,     :))**2
!
! smoothness indicators for negative wind case
!
  ZBNEG1(:,IS-1:IN-1,:) = (PSRC(:,IS:IN,:) - PSRC(:,IS+1:IN+1,:))**2
  ZBNEG1(:,IN,       :) = (PSRC(:,IN+1, :) - TNORTH(:,:))**2
!
  ZBNEG2(:,IS-1:IN,:) = (PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))**2
  ZBNEG2(:,IN+1,   :) = (PSRC(:,IN+1,   :) - TNORTH(:,:))**2
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / (ZEPS + ZBPOS1)**2
  ZOMP2 = ZGAMMA2 / (ZEPS + ZBPOS2)**2
  ZOMN1 = ZGAMMA1 / (ZEPS + ZBNEG1)**2
  ZOMN2 = ZGAMMA2 / (ZEPS + ZBNEG2)**2
#else
!
! smoothness indicators for positive wind case
!
  ZBPOS1(:,IS:IN+1,:) = BR_P2(PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))
  ZBPOS1(:,IS-1,   :) = BR_P2(PSRC(:,IS-1,   :) - TSOUTH(:,:))
!
  ZBPOS2(:,IS-1:IN,:) = BR_P2(PSRC(:,IS:IN+1,:) - PSRC(:,IS-1:IN,:))
  ZBPOS2(:,IN+1,   :) = BR_P2(TNORTH(:,:)       - PSRC(:,IN+1,     :))
!
! smoothness indicators for negative wind case
!
  ZBNEG1(:,IS-1:IN-1,:) = BR_P2(PSRC(:,IS:IN,:) - PSRC(:,IS+1:IN+1,:))
  ZBNEG1(:,IN,       :) = BR_P2(PSRC(:,IN+1, :) - TNORTH(:,:))
!
  ZBNEG2(:,IS-1:IN,:) = BR_P2(PSRC(:,IS-1:IN,:) - PSRC(:,IS:IN+1,:))
  ZBNEG2(:,IN+1,   :) = BR_P2(PSRC(:,IN+1,   :) - TNORTH(:,:))
!
! WENO weights
!
  ZOMP1 = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1)
  ZOMP2 = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2)
  ZOMN1 = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1)
  ZOMN2 = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2)
#endif
!
  PR = (ZOMP2/(ZOMP1+ZOMP2) * ZFPOS2 +                           &
       (ZOMP1/(ZOMP1+ZOMP2) * ZFPOS1)) * (0.5+SIGN(0.5,PRVCT)) + &
       (ZOMN2/(ZOMN1+ZOMN2) * ZFNEG2 +                           &
       (ZOMN1/(ZOMN1+ZOMN2) * ZFNEG1)) * (0.5-SIGN(0.5,PRVCT))
!
!
!       OPEN, WALL, NEST CASE IN THE Y DIRECTION
!
CASE ('OPEN','WALL','NEST')
!
  IS=IJB
  IN=IJE
!
!       USE A FIRST ORDER UPSTREAM SCHEME AT THE PHYSICAL BORDER
!
  IF(LSOUTH_ll()) THEN
    PR(:,IS-1,:) = PSRC(:,IS-1,:) * (0.5+SIGN(0.5,PRVCT(:,IS-1,:))) + PSRC(:,IS,:) * (0.5-SIGN(0.5,PRVCT(:,IS-1,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(:,IS-1,:) = 0.5 * (3.0*PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZFPOS2(:,IS-1,:) = 0.5 * (PSRC(:,IS-1,    :) + PSRC(:,IS,:))
    ZBPOS1(:,IS-1,:) = (PSRC(:,IS-1,:) - TSOUTH(:,:))**2
    ZBPOS2(:,IS-1,:) = (PSRC(:,IS,  :) - PSRC(:,IS-1,:))**2
!
    ZFNEG1(:,IS-1,:) = 0.5 * (3.0*PSRC(:,IS,:) - PSRC(:,IS+1,:))
    ZFNEG2(:,IS-1,:) = 0.5 * (PSRC(:,IS-1,  :) + PSRC(:,IS,:))
    ZBNEG1(:,IS-1,:) = (PSRC(:,IS,:) - PSRC(:,IS+1,:))**2
    ZBNEG2(:,IS-1,:) = (PSRC(:,IS-1,:) - PSRC(:,IS,:))**2
!
    ZOMP1(:,IS-1,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IS-1,:))**2
    ZOMP2(:,IS-1,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IS-1,:))**2
    ZOMN1(:,IS-1,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IS-1,:))**2
    ZOMN2(:,IS-1,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IS-1,:))**2
#else
    ZFPOS1(:,IS-1,:) = 0.5 * (3.0*PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZFPOS2(:,IS-1,:) = 0.5 * (PSRC(:,IS-1,    :) + PSRC(:,IS,:))
    ZBPOS1(:,IS-1,:) = BR_P2(PSRC(:,IS-1,:) - TSOUTH(:,:))
    ZBPOS2(:,IS-1,:) = BR_P2(PSRC(:,IS,  :) - PSRC(:,IS-1,:))
!
    ZFNEG1(:,IS-1,:) = 0.5 * (3.0*PSRC(:,IS,:) - PSRC(:,IS+1,:))
    ZFNEG2(:,IS-1,:) = 0.5 * (PSRC(:,IS-1,  :) + PSRC(:,IS,:))
    ZBNEG1(:,IS-1,:) = BR_P2(PSRC(:,IS,:) - PSRC(:,IS+1,:))
    ZBNEG2(:,IS-1,:) = BR_P2(PSRC(:,IS-1,:) - PSRC(:,IS,:))
!
    ZOMP1(:,IS-1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IS-1,:))
    ZOMP2(:,IS-1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IS-1,:))
    ZOMN1(:,IS-1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IS-1,:))
    ZOMN2(:,IS-1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IS-1,:))
#endif
!
    PR(:,IS-1,:) = (ZOMP2(:,IS-1,:)/(ZOMP1(:,IS-1,:)+ZOMP2(:,IS-1,:)) * ZFPOS2(:,IS-1,:) +   &
       (ZOMP1(:,IS-1,:)/(ZOMP1(:,IS-1,:)+ZOMP2(:,IS-1,:)) * ZFPOS1(:,IS-1,:))) * (0.5+SIGN(0.5,PRVCT(:,IS-1,:))) + &
       (ZOMN2(:,IS-1,:)/(ZOMN1(:,IS-1,:)+ZOMN2(:,IS-1,:)) * ZFNEG2(:,IS-1,:) +        &
       (ZOMN1(:,IS-1,:)/(ZOMN1(:,IS-1,:)+ZOMN2(:,IS-1,:)) * ZFNEG1(:,IS-1,:))) * (0.5-SIGN(0.5,PRVCT(:,IS-1,:)))
!
  ENDIF
!
  IF(LNORTH_ll()) THEN
    PR(:,IN,:) = PSRC(:,IN,:) * (0.5+SIGN(0.5,PRVCT(:,IN,:))) + PSRC(:,IN+1,:) * (0.5-SIGN(0.5,PRVCT(:,IN,:)))
!
!!$  ELSEIF (NHALO == 1) THEN
  ELSE
#ifndef MNH_BITREP
    ZFPOS1(:,IN,:) = 0.5 * (3.0*PSRC(:,IN,:) - PSRC(:,IN-1,:))
    ZFPOS2(:,IN,:) = 0.5 * (PSRC(:,IN,    :) + PSRC(:,IN+1,:))
    ZBPOS1(:,IN,:) = (PSRC(:,IN,  :) - PSRC(:,IN-1,:))**2
    ZBPOS2(:,IN,:) = (PSRC(:,IN+1,:) - PSRC(:,IN,  :))**2
!
    ZFNEG1(:,IN,:) = 0.5 * (3.0*PSRC(:,IN+1,:) - TNORTH(:,:))
    ZFNEG2(:,IN,:) = 0.5 * (PSRC(:,IN,      :) + PSRC(:,IN+1,:))
    ZBNEG1(:,IN,:) = (PSRC(:,IN+1,:) - TNORTH(:,:))**2
    ZBNEG2(:,IN,:) = (PSRC(:,IN,  :) - PSRC(:,IN+1,:))**2
!
    ZOMP1(:,IN,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IN,:))**2
    ZOMP2(:,IN,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IN,:))**2
    ZOMN1(:,IN,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IN,:))**2
    ZOMN2(:,IN,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IN,:))**2
#else
    ZFPOS1(:,IN,:) = 0.5 * (3.0*PSRC(:,IN,:) - PSRC(:,IN-1,:))
    ZFPOS2(:,IN,:) = 0.5 * (PSRC(:,IN,    :) + PSRC(:,IN+1,:))
    ZBPOS1(:,IN,:) = BR_P2(PSRC(:,IN,  :) - PSRC(:,IN-1,:))
    ZBPOS2(:,IN,:) = BR_P2(PSRC(:,IN+1,:) - PSRC(:,IN,  :))
!
    ZFNEG1(:,IN,:) = 0.5 * (3.0*PSRC(:,IN+1,:) - TNORTH(:,:))
    ZFNEG2(:,IN,:) = 0.5 * (PSRC(:,IN,      :) + PSRC(:,IN+1,:))
    ZBNEG1(:,IN,:) = BR_P2(PSRC(:,IN+1,:) - TNORTH(:,:))
    ZBNEG2(:,IN,:) = BR_P2(PSRC(:,IN,  :) - PSRC(:,IN+1,:))
!
    ZOMP1(:,IN,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IN,:))
    ZOMP2(:,IN,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IN,:))
    ZOMN1(:,IN,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IN,:))
    ZOMN2(:,IN,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IN,:))
#endif
!
    PR(:,IN,:) = (ZOMP2(:,IN,:)/(ZOMP1(:,IN,:)+ZOMP2(:,IN,:)) * ZFPOS2(:,IN,:) + &
       (ZOMP1(:,IN,:)/(ZOMP1(:,IN,:)+ZOMP2(:,IN,:)) * ZFPOS1(:,IN,:))) * (0.5+SIGN(0.5,PRVCT(:,IN,:))) + &
       (ZOMN2(:,IN,:)/(ZOMN1(:,IN,:)+ZOMN2(:,IN,:)) * ZFNEG2(:,IN,:) +  &
       (ZOMN1(:,IN,:)/(ZOMN1(:,IN,:)+ZOMN2(:,IN,:)) * ZFNEG1(:,IN,:))) * (0.5-SIGN(0.5,PRVCT(:,IN,:)))
!
  ENDIF
!
!      USE A THIRD ORDER UPSTREAM WENO SCHEME ELSEWHERE 
!
#ifndef MNH_BITREP
  ZFPOS1(:,IS:IN-1,:) = 0.5 * (3.0*PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZFPOS2(:,IS:IN-1,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBPOS1(:,IS:IN-1,:) = (PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))**2
  ZBPOS2(:,IS:IN-1,:) = (PSRC(:,IS+1:IN,:) - PSRC(:,IS:IN-1,  :))**2
!
  ZFNEG1(:,IS:IN-1,:) = 0.5 * (3.0*PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZFNEG2(:,IS:IN-1,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBNEG1(:,IS:IN-1,:) = (PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))**2
  ZBNEG2(:,IS:IN-1,:) = (PSRC(:,IS:IN-1,:) - PSRC(:,IS+1:IN,  :))**2
!
  ZOMP1(:,IS:IN-1,:) = ZGAMMA1 / (ZEPS + ZBPOS1(:,IS:IN-1,:))**2
  ZOMP2(:,IS:IN-1,:) = ZGAMMA2 / (ZEPS + ZBPOS2(:,IS:IN-1,:))**2
  ZOMN1(:,IS:IN-1,:) = ZGAMMA1 / (ZEPS + ZBNEG1(:,IS:IN-1,:))**2
  ZOMN2(:,IS:IN-1,:) = ZGAMMA2 / (ZEPS + ZBNEG2(:,IS:IN-1,:))**2
#else
  ZFPOS1(:,IS:IN-1,:) = 0.5 * (3.0*PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZFPOS2(:,IS:IN-1,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBPOS1(:,IS:IN-1,:) = BR_P2(PSRC(:,IS:IN-1,:) - PSRC(:,IS-1:IN-2,:))
  ZBPOS2(:,IS:IN-1,:) = BR_P2(PSRC(:,IS+1:IN,:) - PSRC(:,IS:IN-1,  :))
!
  ZFNEG1(:,IS:IN-1,:) = 0.5 * (3.0*PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZFNEG2(:,IS:IN-1,:) = 0.5 * (PSRC(:,IS:IN-1,    :) + PSRC(:,IS+1:IN,  :))
  ZBNEG1(:,IS:IN-1,:) = BR_P2(PSRC(:,IS+1:IN,:) - PSRC(:,IS+2:IN+1,:))
  ZBNEG2(:,IS:IN-1,:) = BR_P2(PSRC(:,IS:IN-1,:) - PSRC(:,IS+1:IN,  :))
!
  ZOMP1(:,IS:IN-1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,IS:IN-1,:))
  ZOMP2(:,IS:IN-1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,IS:IN-1,:))
  ZOMN1(:,IS:IN-1,:) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,IS:IN-1,:))
  ZOMN2(:,IS:IN-1,:) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,IS:IN-1,:))
#endif
!
  PR(:,IS:IN-1,:) = (ZOMP2(:,IS:IN-1,:)/(ZOMP1(:,IS:IN-1,:)+ZOMP2(:,IS:IN-1,:)) * ZFPOS2(:,IS:IN-1,:) + &
       (ZOMP1(:,IS:IN-1,:)/(ZOMP1(:,IS:IN-1,:)+ZOMP2(:,IS:IN-1,:)) * ZFPOS1(:,IS:IN-1,:))) * (0.5+SIGN(0.5,PRVCT(:,IS:IN-1,:))) + &
       (ZOMN2(:,IS:IN-1,:)/(ZOMN1(:,IS:IN-1,:)+ZOMN2(:,IS:IN-1,:)) * ZFNEG2(:,IS:IN-1,:) + &
       (ZOMN1(:,IS:IN-1,:)/(ZOMN1(:,IS:IN-1,:)+ZOMN2(:,IS:IN-1,:)) * ZFNEG1(:,IS:IN-1,:))) * (0.5-SIGN(0.5,PRVCT(:,IS:IN-1,:)))
!
END SELECT
!
PR = PR * PRVCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"ADVEC_WENO_K_2_VY end:PR")
END IF

END SUBROUTINE ADVEC_WENO_K_2_VY
!
!-------------------------------------------------------------------------------
!
!     ############################################
#ifndef _OPENACC
      FUNCTION WENO_K_2_WZ(PSRC, PRWCT) RESULT(PR)
#else
      SUBROUTINE WENO_K_2_WZ(PSRC, PRWCT, PR,                                &
                             ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                             ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     ############################################
!!
!!* Computes PRWCT * PWT. Upstream fluxes of W in Z direction.  
!!  Input PWT is on W Grid 'ie' (i,j,k) based on WGRID reference
!!  Output PR is on mass Grid 'ie' (i,j,k+1/2) based on WGRID reference
!!
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*
!!
!!    MODIFICATIONS
!!    -------------
!!
!-------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_CONF
USE MODD_PARAMETERS,ONLY: JPVEXT
USE MODI_GET_HALO
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use mode_mppdb
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on W grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on MASS GRID
!$acc declare present(PSRC,PRWCT)
!
! output source term
#ifndef _OPENACC
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: PR
#else
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IB    ! Begining useful area in x,y,z directions
INTEGER :: IT    ! End useful area in x,y,z directions
!
! WENO-related variables:
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "WENO_K_2_WZ beg:PSRC")
  CALL MPPDB_CHECK(PRWCT,"WENO_K_2_WZ beg:PRWCT")
END IF
!
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
IB = 1 + JPVEXT
IT = SIZE(PSRC,3) - JPVEXT
!
!$acc kernels
!
PR(:,:,:) = 0.0
!
ZFPOS1 = 0.0
ZFPOS2 = 0.0
ZFNEG1 = 0.0
ZFNEG2 = 0.0
ZBPOS1 = 0.0
ZBPOS2 = 0.0
ZBNEG1 = 0.0
ZBNEG2 = 0.0
ZOMP1  = 0.0
ZOMP2  = 0.0
ZOMN1  = 0.0
ZOMN2  = 0.0 
!
! intermediate fluxes at the mass point on Wgrid w(i,j,k+1/2) for positive 
! wind case (L. to the R.)
! (r=1 for the first stencil ZFPOS1, r=0 for the second ZFPOS2)
!
ZFPOS1(:,:,IB:IT-1) = 0.5 * (3.0*PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))
ZFPOS2(:,:,IB:IT-1) = 0.5 * (PSRC(:,:,IB:IT-1) + PSRC(:,:,IB+1:IT))
!
! intermediate flux at the mass point on Wgrid w(i,j,k+1/2) for negative 
! wind case (R. to the L.)
! (r=-1 for the first stencil ZFNEG1, r=0 for the second ZFNEG2=ZFPOS2)
!
ZFNEG1(:,:,IB-1:IT-1) = 0.5 * (3.0*PSRC(:,:,IB:IT) - PSRC(:,:,IB+1:IT+1))
ZFNEG2(:,:,IB-1:IT) = 0.5 * (PSRC(:,:,IB-1:IT) + PSRC(:,:,IB:IT+1))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
ZBPOS1(:,:,IB:IT-1) = (PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))**2
ZBPOS2(:,:,IB:IT-1) = (PSRC(:,:,IB+1:IT) - PSRC(:,:,IB:IT-1))**2
!
! smoothness indicators for negative wind case
!
ZBNEG1(:,:,IB-1:IT-1) = (PSRC(:,:,IB:IT) - PSRC(:,:,IB+1:IT+1))**2
ZBNEG2(:,:,IB-1:IT) = (PSRC(:,:,IB-1:IT) - PSRC(:,:,IB:IT+1))**2
!
! WENO weights
!
ZOMP1 = ZGAMMA1 / (ZEPS + ZBPOS1)**2
ZOMP2 = ZGAMMA2 / (ZEPS + ZBPOS2)**2
ZOMN1 = ZGAMMA1 / (ZEPS + ZBNEG1)**2
ZOMN2 = ZGAMMA2 / (ZEPS + ZBNEG2)**2
#else
!
! smoothness indicators for positive wind case
!
ZBPOS1(:,:,IB:IT-1) = BR_P2(PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))
ZBPOS2(:,:,IB:IT-1) = BR_P2(PSRC(:,:,IB+1:IT) - PSRC(:,:,IB:IT-1))
!
! smoothness indicators for negative wind case
!
ZBNEG1(:,:,IB-1:IT-1) = BR_P2(PSRC(:,:,IB:IT) - PSRC(:,:,IB+1:IT+1))
ZBNEG2(:,:,IB-1:IT) = BR_P2(PSRC(:,:,IB-1:IT) - PSRC(:,:,IB:IT+1))
!
! WENO weights
!
ZOMP1 = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1)
ZOMP2 = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2)
ZOMN1 = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1)
ZOMN2 = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2)
#endif
!
! WENO fluxes
!
PR(:,:,IB:IT-1) = (ZOMP2(:,:,IB:IT-1)/(ZOMP1(:,:,IB:IT-1)+ZOMP2(:,:,IB:IT-1))* &
                                                         ZFPOS2(:,:,IB:IT-1) + &
                  (ZOMP1(:,:,IB:IT-1)/(ZOMP1(:,:,IB:IT-1)+ZOMP2(:,:,IB:IT-1))* &
                 ZFPOS1(:,:,IB:IT-1))) * (0.5+SIGN(0.5,PRWCT(:,:,IB:IT-1) )) + &
                  (ZOMN2(:,:,IB:IT-1)/(ZOMN1(:,:,IB:IT-1)+ZOMN2(:,:,IB:IT-1))* &
                                                         ZFNEG2(:,:,IB:IT-1) + &
                  (ZOMN1(:,:,IB:IT-1)/(ZOMN1(:,:,IB:IT-1)+ZOMN2(:,:,IB:IT-1))* &
                 ZFNEG1(:,:,IB:IT-1))) * (0.5-SIGN(0.5,PRWCT(:,:,IB:IT-1) ))
!
PR(:,:,IB-1) = PSRC(:,:,IB-1) * (0.5+SIGN(0.5,PRWCT(:,:,IB-1) )) + &
               PSRC(:,:,IB)   * (0.5-SIGN(0.5,PRWCT(:,:,IB-1) ))
PR(:,:,IT)   = PSRC(:,:,IT)   * (0.5+SIGN(0.5,PRWCT(:,:,IT) ))   + &
               PSRC(:,:,IT+1) * (0.5-SIGN(0.5,PRWCT(:,:,IT) ))
PR(:,:,IT+1) = -999.
!
PR = PR * PRWCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"WENO_K_2_WZ end:PR")
END IF

#ifndef _OPENACC
END FUNCTION WENO_K_2_WZ
#else
END SUBROUTINE WENO_K_2_WZ
#endif
!
!-----------------------------------------------------------------------------
!
!     ############################################
#ifndef _OPENACC
      FUNCTION WENO_K_2_MZ(PSRC, PRWCT) RESULT(PR)
#else
      SUBROUTINE WENO_K_2_MZ(PSRC, PRWCT, PR,                                &
                             ZFPOS1, ZFPOS2, ZFNEG1, ZFNEG2, ZBPOS1, ZBPOS2, &
                             ZBNEG1, ZBNEG2, ZOMP1, ZOMP2, ZOMN1, ZOMN2)
#endif
!     ############################################
!!
!!* Computes PRWCT * PUT (or PRWCT * PVT). Upstream fluxes of U (or V) 
!!  variables in Z direction.  
!!  Input PUT is on U Grid 'ie' (i,j,k) based on UGRID reference                
!!  Output PR is on mass Grid 'ie' (i,j,k-1/2) based on UGRID reference
!!
!!    AUTHOR
!!    ------
!!    F. Visentin   *CNRS/LA*
!!
!!    MODIFICATIONS
!!    -------------
!!
!-------------------------------------------------------------------------------
!
USE MODE_ll
USE MODD_CONF
USE MODD_PARAMETERS,ONLY: JPVEXT
USE MODI_GET_HALO
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use mode_mppdb
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC  ! variable on MASS grid at t
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRWCT ! contrav. comp. on W grid
!$acc declare present(PSRC,PRWCT)
!
! output source term
!
#ifndef _OPENACC
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: PR
#else
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PR
!$acc declare present(PR)
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER :: IB    ! Begining useful area in x,y,z directions
INTEGER :: IT    ! End useful area in x,y,z directions
!
! WENO-related variables:
!
! intermediate reconstruction fluxes for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFPOS1, ZFPOS2
!$acc declare present(ZFPOS1,ZFPOS2)
!
! intermediate reconstruction fluxes for negative wind case
! we need only one since ZFNEG2 = ZFPOS2
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZFNEG1, ZFNEG2
!$acc declare present(ZFNEG1,ZFNEG2)
!
! smoothness indicators for positive wind case
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBPOS1, ZBPOS2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZBNEG1, ZBNEG2
!$acc declare present(ZBPOS1,ZBPOS2,ZBNEG1,ZBNEG2)
!
! WENO weights
!
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMP1, ZOMP2
REAL, DIMENSION(SIZE(PSRC,1),SIZE(PSRC,2),SIZE(PSRC,3)) :: ZOMN1, ZOMN2
!$acc declare present(ZOMP1,ZOMP2,ZOMN1,ZOMN2)
!
! standard weights
!
REAL, PARAMETER :: ZGAMMA1 = 1./3.
REAL, PARAMETER :: ZGAMMA2 = 2./3.
!
REAL, PARAMETER :: ZEPS = 1.0E-15
!
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PSRC, "WENO_K_2_MZ beg:PSRC")
  CALL MPPDB_CHECK(PRWCT,"WENO_K_2_MZ beg:PRWCT")
END IF
!
!*       0.3.     COMPUTES THE DOMAIN DIMENSIONS
!                 ------------------------------
!
IB = 1 + JPVEXT
IT = SIZE(PSRC,3) - JPVEXT
!
!$acc kernels
!
PR(:,:,:) = 0.0
!
ZFPOS1 = 0.0
ZFPOS2 = 0.0
ZFNEG1 = 0.0
ZFNEG2 = 0.0 
ZBPOS1 = 0.0
ZBPOS2 = 0.0
ZBNEG1 = 0.0
ZBNEG2 = 0.0
ZOMP1  = 0.0
ZOMP2  = 0.0
ZOMN1  = 0.0
ZOMN2  = 0.0
!
! intermediate fluxes at the flux point on the Wgrid u(i,j,k-1/2) for 
! positive wind case
!
ZFPOS1(:,:,IB+1:IT) = 0.5 * (3.0*PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))
ZFPOS2(:,:,IB+1:IT) = 0.5 * (PSRC(:,:,IB:IT-1) + PSRC(:,:,IB+1:IT))
!
! intermediate flux at the flux point on the Wgrid u(i,j,k-1/2) for 
! negative wind case
!
ZFNEG1(:,:,IB+1:IT) = 0.5 * (3.0*PSRC(:,:,IB+1:IT) - PSRC(:,:,IB+2:IT+1))
ZFNEG2(:,:,IB+1:IT) = 0.5 * (PSRC(:,:,IB:IT-1) + PSRC(:,:,IB+1:IT))
#ifndef MNH_BITREP
!
! smoothness indicators for positive wind case
!
ZBPOS1(:,:,IB+1:IT) = (PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))**2
ZBPOS2(:,:,IB+1:IT) = (PSRC(:,:,IB+1:IT) - PSRC(:,:,IB:IT-1))**2
!
! smoothness indicators for negative wind case
!
ZBNEG1(:,:,IB+1:IT) = (PSRC(:,:,IB+1:IT) - PSRC(:,:,IB+2:IT+1))**2
ZBNEG2(:,:,IB+1:IT) = (PSRC(:,:,IB:IT-1) - PSRC(:,:,IB+1:IT))**2
!
! WENO weights
!
ZOMP1(:,:,IB+1:IT) = ZGAMMA1 / (ZEPS + ZBPOS1(:,:,IB+1:IT))**2
ZOMP2(:,:,IB+1:IT) = ZGAMMA2 / (ZEPS + ZBPOS2(:,:,IB+1:IT))**2
ZOMN1(:,:,IB+1:IT) = ZGAMMA1 / (ZEPS + ZBNEG1(:,:,IB+1:IT))**2
ZOMN2(:,:,IB+1:IT) = ZGAMMA2 / (ZEPS + ZBNEG2(:,:,IB+1:IT))**2
#else
!
! smoothness indicators for positive wind case
!
ZBPOS1(:,:,IB+1:IT) = BR_P2(PSRC(:,:,IB:IT-1) - PSRC(:,:,IB-1:IT-2))
ZBPOS2(:,:,IB+1:IT) = BR_P2(PSRC(:,:,IB+1:IT) - PSRC(:,:,IB:IT-1))
!
! smoothness indicators for negative wind case
!
ZBNEG1(:,:,IB+1:IT) = BR_P2(PSRC(:,:,IB+1:IT) - PSRC(:,:,IB+2:IT+1))
ZBNEG2(:,:,IB+1:IT) = BR_P2(PSRC(:,:,IB:IT-1) - PSRC(:,:,IB+1:IT))
!
! WENO weights
!
ZOMP1(:,:,IB+1:IT) = ZGAMMA1 / BR_P2(ZEPS + ZBPOS1(:,:,IB+1:IT))
ZOMP2(:,:,IB+1:IT) = ZGAMMA2 / BR_P2(ZEPS + ZBPOS2(:,:,IB+1:IT))
ZOMN1(:,:,IB+1:IT) = ZGAMMA1 / BR_P2(ZEPS + ZBNEG1(:,:,IB+1:IT))
ZOMN2(:,:,IB+1:IT) = ZGAMMA2 / BR_P2(ZEPS + ZBNEG2(:,:,IB+1:IT))
#endif
!
PR(:,:,IB+1:IT) = (ZOMP2(:,:,IB+1:IT)/(ZOMP1(:,:,IB+1:IT)+ZOMP2(:,:,IB+1:IT))* &
                                                         ZFPOS2(:,:,IB+1:IT) + &
                  (ZOMP1(:,:,IB+1:IT)/(ZOMP1(:,:,IB+1:IT)+ZOMP2(:,:,IB+1:IT))* &
                 ZFPOS1(:,:,IB+1:IT))) * (0.5+SIGN(0.5,PRWCT(:,:,IB+1:IT) )) + &
                  (ZOMN2(:,:,IB+1:IT)/(ZOMN1(:,:,IB+1:IT)+ZOMN2(:,:,IB+1:IT))* &
                                                         ZFNEG2(:,:,IB+1:IT) + &
                  (ZOMN1(:,:,IB+1:IT)/(ZOMN1(:,:,IB+1:IT)+ZOMN2(:,:,IB+1:IT))* &
                 ZFNEG1(:,:,IB+1:IT))) * (0.5-SIGN(0.5,PRWCT(:,:,IB+1:IT) ))
!
PR(:,:,IB)   = PSRC(:,:,IB-1) * (0.5+SIGN(0.5,PRWCT(:,:,IB) ))   + &
               PSRC(:,:,IB)   * (0.5-SIGN(0.5,PRWCT(:,:,IB) ))
PR(:,:,IT+1) = PSRC(:,:,IT)   * (0.5+SIGN(0.5,PRWCT(:,:,IT+1) )) + &
               PSRC(:,:,IT+1) * (0.5-SIGN(0.5,PRWCT(:,:,IT+1) ))
!
PR = PR * PRWCT
!$acc end kernels
!
#ifndef _OPENACC
CALL GET_HALO(PR)
#else
!!$!$acc update self(PR)
CALL GET_HALO_D(PR)
!!$!$acc update device(PR)
#endif
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PR,"WENO_K_2_MZ end:PR")
END IF

#ifndef _OPENACC
END FUNCTION WENO_K_2_MZ
#else
END SUBROUTINE WENO_K_2_MZ
#endif
