#ifdef _OPENACC
!MNH_LIC Copyright 2013-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 10/07/2019: bugfix: MNH_REL_ZT3D_N0: access outside of array was possible
!-----------------------------------------------------------------
MODULE MODE_MNH_ZWORK

  use mode_msg

  IMPLICIT NONE

  INTEGER, SAVE :: IIB,IJB,IKB  ! Begining useful area in x,y,z directions
  INTEGER, SAVE :: IIE,IJE,IKE  ! End useful area in x,y,z directions

  !
  INTEGER,SAVE                    :: IJS,IJN, IIW,IIA
  !
  INTEGER, SAVE                   :: IIU,IJU,IKU
  LOGICAL, SAVE                   :: GWEST , GEAST
  LOGICAL, SAVE                   :: GSOUTH , GNORTH

  LOGICAL, SAVE                   :: GFIRST_CALL_MNH_ALLOC_ZWORK = .TRUE.
  !
  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:)   :: ZPSRC_HALO2_WEST
  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:)   :: ZPSRC_HALO2_SOUTH
!acc declare create(ZPSRC_HALO2_WEST,ZPSRC_HALO2_SOUTH)

  REAL, SAVE, ALLOCATABLE , DIMENSION(:,:,:) :: ZUNIT3D
!acc declare device_resident(ZUNIT3D)

  INTEGER, parameter                        :: JPMAX_T3D = 40
  INTEGER , ALLOCATABLE, DIMENSION (:)      :: NT3D_POOL
  INTEGER                                   :: NT3D_TOP , NT3D_TOP_MAX = 0  
  !REAL    , ALLOCATABLE, DIMENSION(:,:,:,:) , TARGET :: ZT3D_A1,ZT3D_A2,ZT3D_A3,ZT3D_A4
  !REAL    , POINTER    , DIMENSION(:,:,:,:)          :: ZT3D
  REAL,SAVE    , ALLOCATABLE, DIMENSION(:,:,:,:)          :: ZT3D
  ! acc declare create(ZT3D_A1,ZT3D_A2,ZT3D_A3,ZT3D_A4)
!acc declare device_resident(ZT3D)
  TYPE TMODEL
      REAL    , POINTER, DIMENSION(:,:,:,:) :: X
  END TYPE TMODEL

  TYPE(TMODEL) , DIMENSION(10) :: MODEL

CONTAINS

  SUBROUTINE MNH_ALLOC_ZWORK(IMODEL)

    USE MODE_TOOLS_ll, ONLY   : GET_DIM_EXT_ll, GET_INDICE_ll, LWEST_ll,LEAST_ll, LSOUTH_ll, LNORTH_ll

    USE MODD_PARAMETERS, ONLY : JPVEXT, XUNDEF
    USE MODD_DIM_n,      ONLY : NKMAX

    IMPLICIT NONE

    INTEGER :: IMODEL

    INTEGER :: JI

    IF (GFIRST_CALL_MNH_ALLOC_ZWORK) THEN
       GFIRST_CALL_MNH_ALLOC_ZWORK = .FALSE.
       !
       ! Array dim
       !
       CALL GET_DIM_EXT_ll('B',IIU,IJU)     
       IKU=NKMAX + 2* JPVEXT
       !
       ! Computation bound
       !
       CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
       IJS=IJB
       IJN=IJE
       IIW=IIB
       IIA=IIE
       IKB = 1     + JPVEXT
       IKE = NKMAX + JPVEXT
       !
       ! Lateral boundary
       !
       GWEST = LWEST_ll()
       GEAST = LEAST_ll()

       GSOUTH=LSOUTH_ll()
       GNORTH=LNORTH_ll()

       !
       ! Work array
       ! 
       ALLOCATE (ZPSRC_HALO2_WEST(IJU,IKU)) 
       ALLOCATE (ZPSRC_HALO2_SOUTH(IIU,IKU)) 
       !$acc enter data create(ZPSRC_HALO2_WEST,ZPSRC_HALO2_SOUTH)

       ALLOCATE (ZUNIT3D(IIU,IJU,IKU))
       !$acc enter data create(ZUNIT3D)

       !ALLOCATE (ZT3D_A1(IIU,IJU,IKU,JPMAX_T3D))
       !MODEL(1)%X => ZT3D_A1
       !ZT3D => MODEL(1)%X
       ALLOCATE (ZT3D(IIU,IJU,IKU,JPMAX_T3D))
       !$acc enter data create(ZT3D)

       ALLOCATE (NT3D_POOL(JPMAX_T3D))
       NT3D_TOP = 0
       DO JI = 1, JPMAX_T3D
          NT3D_POOL(JI) = JI
       END DO

       !$acc kernels 

       ZPSRC_HALO2_WEST = XUNDEF
       ZPSRC_HALO2_SOUTH = XUNDEF

       ZUNIT3D = 1.0

       ZT3D  =  XUNDEF

       !$acc end kernels

       !$acc update host (ZPSRC_HALO2_WEST,ZPSRC_HALO2_SOUTH)
       !$acc update host (ZUNIT3D)
       !$acc update host (ZT3D)

    END IF

  END SUBROUTINE MNH_ALLOC_ZWORK

  SUBROUTINE MNH_GET_ZT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF (NT3D_TOP == JPMAX_T3D ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT3D_N0', 'NT3D_TOP too big (increaze JPMAX_T3D)' )
    ELSE
       NT3D_TOP = NT3D_TOP + 1
       KTEMP = NT3D_POOL(NT3D_TOP)
       IF (  NT3D_POOL(NT3D_TOP) == -1 ) THEN
         call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT3D_N0', 'slice already reserved' )
       END IF
       NT3D_POOL(NT3D_TOP) = -1
       IF (  NT3D_TOP > NT3D_TOP_MAX ) THEN
          NT3D_TOP_MAX = NT3D_TOP
          WRITE( *, '( " MNH_GET_ZT3D: NT3D_TOP_MAX=",I4," KTEMP=",I4 )' ) NT3D_TOP_MAX,KTEMP
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_ZT3D: reserving ZT3D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_GET_ZT3D_N0

  SUBROUTINE MNH_GET_ZT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_GET_ZT3D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_GET_ZT3D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_GET_ZT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_GET_ZT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_GET_ZT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_GET_ZT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_GET_ZT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_GET_ZT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_GET_ZT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_GET_ZT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_GET_ZT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_GET_ZT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_GET_ZT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_GET_ZT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_GET_ZT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_GET_ZT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_GET_ZT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_GET_ZT3D_N0(KTEMP18)


  END SUBROUTINE MNH_GET_ZT3D

  SUBROUTINE MNH_GET_ZT4D(KSIZE,KBEG,KEND)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: KSIZE
    INTEGER, INTENT(OUT) :: KBEG, KEND

    INTEGER :: JI

    IF (NT3D_TOP + KSIZE > JPMAX_T3D ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT4D', 'NT3D_TOP too big (increaze JPMAX_T3D)' )
    ELSE
       KBEG = NT3D_TOP + 1
       KEND = NT3D_TOP + KSIZE
       NT3D_TOP = NT3D_TOP + KSIZE
       DO JI = KBEG, KEND
         IF (NT3D_POOL(JI) == -1) THEN
           call Print_msg( NVERB_FATAL, 'GEN', 'MNH_GET_ZT4D', 'trying to use area already reserved' )
         END IF
         NT3D_POOL(JI) = -1
       END DO
       IF ( NT3D_TOP > NT3D_TOP_MAX ) THEN
          NT3D_TOP_MAX = NT3D_TOP
          WRITE( *, '( " MNH_GET_ZT4D: NT3D_TOP_MAX=",I4," KBEG=",I4," KEND=",I4 )' ) NT3D_TOP_MAX,KBEG,KEND
       END IF
    ENDIF
    !WRITE( *, '( "MNH_GET_ZT4D: reserving ZT3D (",I4,I4,")" )' ) KBEG,KEND

  END SUBROUTINE MNH_GET_ZT4D

  SUBROUTINE MNH_REL_ZT3D_N0(KTEMP)

    IMPLICIT NONE

    INTEGER :: KTEMP

    IF ( ( NT3D_TOP > JPMAX_T3D ) .OR. ( NT3D_TOP < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT3D_N0', 'invalid value for NT3D_TOP' )
    ELSE
      NT3D_POOL(KTEMP) = KTEMP
      IF (KTEMP == NT3D_TOP) THEN
        NT3D_TOP = NT3D_TOP - 1
        DO WHILE (NT3D_TOP > 0 )
          if ( NT3D_POOL(NT3D_TOP) == -1 ) exit
          NT3D_TOP = NT3D_TOP - 1
        END DO
      END IF
    ENDIF
    !WRITE( *, '( "MNH_REL_ZT3D: releasing ZT3D (",I4,")" )' ) KTEMP

  END SUBROUTINE MNH_REL_ZT3D_N0

  SUBROUTINE MNH_REL_ZT3D(KTEMP1,KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9, &
                          KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18)

    IMPLICIT NONE

    INTEGER          :: KTEMP1
    INTEGER,OPTIONAL :: KTEMP2,KTEMP3,KTEMP4,KTEMP5,KTEMP6,KTEMP7,KTEMP8,KTEMP9
    INTEGER,OPTIONAL :: KTEMP10,KTEMP11,KTEMP12,KTEMP13,KTEMP14,KTEMP15,KTEMP16,KTEMP17,KTEMP18

                          CALL MNH_REL_ZT3D_N0(KTEMP1)
    IF (PRESENT(KTEMP2))  CALL MNH_REL_ZT3D_N0(KTEMP2)
    IF (PRESENT(KTEMP3))  CALL MNH_REL_ZT3D_N0(KTEMP3)
    IF (PRESENT(KTEMP4))  CALL MNH_REL_ZT3D_N0(KTEMP4)
    IF (PRESENT(KTEMP5))  CALL MNH_REL_ZT3D_N0(KTEMP5)
    IF (PRESENT(KTEMP6))  CALL MNH_REL_ZT3D_N0(KTEMP6)
    IF (PRESENT(KTEMP7))  CALL MNH_REL_ZT3D_N0(KTEMP7)
    IF (PRESENT(KTEMP8))  CALL MNH_REL_ZT3D_N0(KTEMP8)
    IF (PRESENT(KTEMP9))  CALL MNH_REL_ZT3D_N0(KTEMP9)
    IF (PRESENT(KTEMP10)) CALL MNH_REL_ZT3D_N0(KTEMP10)
    IF (PRESENT(KTEMP11)) CALL MNH_REL_ZT3D_N0(KTEMP11)
    IF (PRESENT(KTEMP12)) CALL MNH_REL_ZT3D_N0(KTEMP12)
    IF (PRESENT(KTEMP13)) CALL MNH_REL_ZT3D_N0(KTEMP13)
    IF (PRESENT(KTEMP14)) CALL MNH_REL_ZT3D_N0(KTEMP14)
    IF (PRESENT(KTEMP15)) CALL MNH_REL_ZT3D_N0(KTEMP15)
    IF (PRESENT(KTEMP16)) CALL MNH_REL_ZT3D_N0(KTEMP16)
    IF (PRESENT(KTEMP17)) CALL MNH_REL_ZT3D_N0(KTEMP17)
    IF (PRESENT(KTEMP18)) CALL MNH_REL_ZT3D_N0(KTEMP18)

  END SUBROUTINE MNH_REL_ZT3D

  SUBROUTINE MNH_REL_ZT4D(KSIZE,KBEG)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: KSIZE
    INTEGER, INTENT(IN) :: KBEG

    character(len=16) :: ytxt1, ytxt2
    INTEGER :: JI

    IF ( KBEG + KSIZE -1 /= NT3D_TOP ) THEN
!PW TODO: implement holes management
      write( ytxt1, '( I4, "-", I4 )' ) kbeg, kbeg + ksize - 1
      write( ytxt2, '( I4 )' )          NT3D_TOP
      call Print_msg( NVERB_ERROR, 'GEN', 'MNH_REL_ZT4D', 'trying to free area (' // trim( ytxt1 ) // &
                                                          ') not at the end of range (' // trim( ytxt2 ) // ')' )
    END IF
    IF ( ( KBEG + KSIZE > JPMAX_T3D ) .OR. ( KBEG < 1 ) ) THEN
      call Print_msg( NVERB_FATAL, 'GEN', 'MNH_REL_ZT4D', 'out of range' )
    END IF

    DO JI = KBEG, KBEG+KSIZE-1
      IF (NT3D_POOL(JI) /= -1) THEN
      call Print_msg( NVERB_ERROR, 'GEN', 'MNH_REL_ZT4D', 'trying to free area not reserved' )
      END IF
      NT3D_POOL(JI) = JI
    END DO
    NT3D_TOP = NT3D_TOP - KSIZE
    !WRITE( *, '( "MNH_REL_ZT4D: releasing ZT3D (",I4,I4,")" )' ) KBEG,KBEG+KSIZE-1

  END SUBROUTINE MNH_REL_ZT4D



END MODULE MODE_MNH_ZWORK
#endif
