!MNH_LIC Copyright 2006-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-------------------------------------------------------------------------------
!
!
!     #####################
      MODULE MODI_PPM_SCALAR
!     #####################
!
INTERFACE
!
      SUBROUTINE PPM_SCALAR (HLBCX,HLBCY, KSV, TPDTCUR,     &
                     PCRU, PCRV, PCRW, PTSTEP, PTSTEP_PPM,  &
                     PRHODJ, PRHOX1, PRHOX2, PRHOY1, PRHOY2,&
                     PRHOZ1, PRHOZ2,                        &
                     PSVT, PRSVS, HSV_ADV_SCHEME            ) 
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODD_TYPE_DATE,   ONLY : DATE_TIME
!
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
CHARACTER (LEN=6),               INTENT(IN) :: HSV_ADV_SCHEME
!
INTEGER,                  INTENT(IN)    :: KSV    ! Number of Scalar Variables
TYPE (DATE_TIME),         INTENT(IN)    :: TPDTCUR ! current date and time
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRU  ! Courant
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRV  ! numbers
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRW  ! 
!$acc declare present(PCRU,PCRV,PCRW)
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ ! density
!$acc declare present(PRHODJ)
! Temporary advected rhodj
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOX1,PRHOX2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOY1,PRHOY2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOZ1,PRHOZ2
!$acc declare present(PRHOX1,PRHOX2,PRHOY1,PRHOY2,PRHOZ1,PRHOZ2)
!
REAL,                     INTENT(IN)    :: PTSTEP ! Time step model  
REAL,                     INTENT(IN)    :: PTSTEP_PPM ! Time Step PPM
!
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PSVT         ! Vars at t
!PW: bug workaround with PGI (tested up to 16.10)
!PSVT is present but if zero-size => not detected as present
!!$acc declare present(PSVT)
!$acc declare pcopyin(PSVT)
!
REAL, DIMENSION(:,:,:,:), INTENT(OUT)   :: PRSVS ! Source terms
!PW: bug workaround with PGI (tested up to 16.10)
!PRSVS is present but if zero-size => not detected as present
!!$acc declare present(PRSVS)
!$acc declare pcopyout(PRSVS)
!
!
END SUBROUTINE PPM_SCALAR
!
END INTERFACE
!
END MODULE MODI_PPM_SCALAR
!
!     ######################################################################
#ifdef _OPENACC
      SUBROUTINE PPM_SCALAR (HLBCX,HLBCY, KSV, TPDTCUR,     &
                     PCRU, PCRV, PCRW, PTSTEP, PTSTEP_PPM,  &
                     PRHODJ, PRHOX1, PRHOX2, PRHOY1, PRHOY2,&
                     PRHOZ1, PRHOZ2,                        &
                     PSVT, PRSVS, HSV_ADV_SCHEME            ) 
  USE MODD_TYPE_DATE, ONLY : DATE_TIME
  USE MODE_MNH_ZWORK, ONLY : ZT3D, MNH_GET_ZT3D , MNH_REL_ZT3D
  IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
CHARACTER (LEN=6),               INTENT(IN) :: HSV_ADV_SCHEME
!
INTEGER,                  INTENT(IN)    :: KSV    ! Number of Scalar Variables
TYPE (DATE_TIME),         INTENT(IN)    :: TPDTCUR ! current date and time
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRU  ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRV  !  components
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRW  ! of momentum
!$acc declare present(PCRU,PCRV,PCRW)
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ ! density
!$acc declare present(PRHODJ)
! Temporary advected rhodj
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOX1,PRHOX2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOY1,PRHOY2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOZ1,PRHOZ2
!$acc declare present(PRHOX1,PRHOX2,PRHOY1,PRHOY2,PRHOZ1,PRHOZ2)
!
REAL,                     INTENT(IN)    :: PTSTEP ! Time step 
REAL,                     INTENT(IN)    :: PTSTEP_PPM ! Time Step PPM
!
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PSVT             
!PW: bug workaround with PGI (tested up to 16.10)
!PSVT is present but if zero-size => not detected as present
!!$acc declare present(PSVT)
!$acc declare pcopyin(PSVT)
!
REAL, DIMENSION(:,:,:,:), INTENT(OUT) :: PRSVS         ! Source terms
!PW: bug workaround with PGI (tested up to 16.10)
!PRSVS is present but if zero-size => not detected as present
!!$acc declare present(PRSVS)
!$acc declare pcopyout(PRSVS)

         CALL  PPM_SCALAR_D (HLBCX,HLBCY, KSV, TPDTCUR,             &
                             PCRU, PCRV, PCRW, PTSTEP, PTSTEP_PPM,  &
                             PRHODJ, PRHOX1, PRHOX2, PRHOY1, PRHOY2,&
                             PRHOZ1, PRHOZ2,                        &
                             PSVT, PRSVS, HSV_ADV_SCHEME)

CONTAINS

      SUBROUTINE PPM_SCALAR_D (HLBCX,HLBCY, KSV, TPDTCUR,  &
                     PCRU, PCRV, PCRW, PTSTEP, PTSTEP_PPM,  &
                     PRHODJ, PRHOX1, PRHOX2, PRHOY1, PRHOY2,&
                     PRHOZ1, PRHOZ2,                     &
                     PSVT, PRSVS, HSV_ADV_SCHEME)
#else
      SUBROUTINE PPM_SCALAR (HLBCX,HLBCY, KSV, TPDTCUR,     &
                     PCRU, PCRV, PCRW, PTSTEP, PTSTEP_PPM,  &
                     PRHODJ, PRHOX1, PRHOX2, PRHOY1, PRHOY2,&
                     PRHOZ1, PRHOZ2,                        &
                     PSVT, PRSVS, HSV_ADV_SCHEME            ) 
#endif
!     ######################################################################
!
!!****  *PPM_SCALAR * 
!!
!!    PURPOSE
!!    -------
!!
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!
!!    MODULE MODD_ARGSLIST
!!         HALO2LIST_ll : type for a list of "HALO2_lls"
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!
!!    MODIFICATIONS
!!    -------------
!!      Original 11.05.2006. T.Maric
!!      Modification : 11.2011 C.Lac, V.Masson : Advection of (theta_l,r_t) 
!!                  10/2016  (C.Lac) Correction on the flag for Strang splitting
!!                                  to insure reproducibility between START and RESTA!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_ARGSLIST_ll, ONLY: HALO2LIST_ll
USE MODD_CONF
USE MODD_PARAMETERS
USE MODD_TYPE_DATE,   ONLY: DATE_TIME
!
use mode_mppdb
#ifdef _OPENACC
use mode_msg
#endif
!
USE MODI_ADVEC_PPM_ALGO
USE MODI_PPM
USE MODI_SHUMAN
!
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
CHARACTER (LEN=6),               INTENT(IN) :: HSV_ADV_SCHEME
!
INTEGER,                  INTENT(IN)    :: KSV    ! Number of Scalar Variables
TYPE (DATE_TIME),         INTENT(IN)    :: TPDTCUR ! current date and time
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRU  ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRV  !  components
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PCRW  ! of momentum
!$acc declare present(PCRU,PCRV,PCRW)
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODJ ! density
!$acc declare present(PRHODJ)
! Temporary advected rhodj
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOX1,PRHOX2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOY1,PRHOY2
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHOZ1,PRHOZ2
!$acc declare present(PRHOX1,PRHOX2,PRHOY1,PRHOY2,PRHOZ1,PRHOZ2)
!
REAL,                     INTENT(IN)    :: PTSTEP ! Time step model  
REAL,                     INTENT(IN)    :: PTSTEP_PPM ! Time Step PPM
!
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PSVT             
!$acc declare present(PSVT)
!
REAL, DIMENSION(:,:,:,:), INTENT(OUT)   :: PRSVS         ! Source terms
!$acc declare present(PRSVS)
!
!
!*       0.2   Declarations of local variables :
!
INTEGER :: JSV           ! Loop index for Scalar Variables
!
INTEGER :: IGRID ! localisation on the model grid
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PCRU,"PPM_SCALAR beg:PCRU")
  CALL MPPDB_CHECK(PCRV,"PPM_SCALAR beg:PCRV")
  CALL MPPDB_CHECK(PCRW,"PPM_SCALAR beg:PCRW")
  CALL MPPDB_CHECK(PRHODJ,"PPM_SCALAR beg:PRHODJ")
  CALL MPPDB_CHECK(PRHOX1,"PPM_SCALAR beg:PRHOX1")
  CALL MPPDB_CHECK(PRHOX2,"PPM_SCALAR beg:PRHOX2")
  CALL MPPDB_CHECK(PRHOY1,"PPM_SCALAR beg:PRHOY1")
  CALL MPPDB_CHECK(PRHOY2,"PPM_SCALAR beg:PRHOY2")
  CALL MPPDB_CHECK(PRHOZ1,"PPM_SCALAR beg:PRHOZ1")
  CALL MPPDB_CHECK(PRHOZ2,"PPM_SCALAR beg:PRHOZ2")
  CALL MPPDB_CHECK(PSVT,"PPM_SCALAR beg:PSVT")
END IF
!
!*       1.     CALL THE ADVEC_PPM_ALGO ROUTINE FOR EACH FIELD
!               -----------------------------------------------
!
IGRID = 1
!
! Case with KSV tracers
!
DO JSV=1,KSV
#ifdef _OPENACC
  call Print_msg( NVERB_WARNING, 'GEN', 'PPM_SCALAR', 'KSV>0 not yet tested' )
#endif
   CALL ADVEC_PPM_ALGO(HSV_ADV_SCHEME, HLBCX, HLBCY, IGRID, PSVT(:,:,:,JSV), & 
                       PRHODJ, PTSTEP, PTSTEP_PPM,                           & 
                       PRHOX1, PRHOX2, PRHOY1, PRHOY2, PRHOZ1, PRHOZ2,       &
                       PRSVS(:,:,:,JSV), TPDTCUR, PCRU, PCRV, PCRW)
END DO
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRSVS,"PPM_SCALAR end:PRSVS")
END IF
!
#ifdef _OPENACC
END SUBROUTINE PPM_SCALAR_D
#endif
!
END SUBROUTINE PPM_SCALAR
